"""Model classes for PyQt's model view architecture.

Copyright (C) 2021  Andreas Hellerschmied <andreas.hellerschmied@bev.gv.at>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from PyQt5.QtCore import QAbstractTableModel, Qt, QPersistentModelIndex
from PyQt5 import QtGui
from PyQt5.QtWidgets import QMessageBox, QStyledItemDelegate
import datetime as dt
import numpy as np

from gravtools.models.survey import Survey

NONE_REPRESENTATION_IN_TABLE_VIEW = ''  # Representation of None values in table views in the GUI


class ObservationTableModel(QAbstractTableModel):
    """Model for displaying the observation data.

    Attributes
    ----------
    _surveys : dict of py.obj:`gravtools.survey.Survey`
        Dict of survey object that are assigned by reference to the campaign data object
        (py.obj:`gravtools.survey.Campaign`).
    _data : pandas dataframe
        Data frame that holds the observation data of a specific setup or survey to be displayed in the table view.
    _data_survey_name : str
        Name of the currently selected survey.
    _setup_data : pandas DataFrame
        Contains one observation per setup.
    """

    _DECIMAL_PLACES_PER_FLOAT_COLUMN = {
        'lon_deg': 3,
        'lat_deg': 3,
        'alt_m': 3,
        'g_obs_mugal': 1,
        'sd_g_obs_mugal': 1,
        'g_red_mugal': 1,
        'sd_g_red_mugal': 1,
        'corr_terrain': 1,
        'corr_tide_mugal': 1,
        'corr_tide_red_mugal': 1,
        'temp': 1,
        'tiltx': 1,
        'tilty': 1,
        'dhf_m': 3,
        'dhb_m': 3,
        'vg_mugalm': 1,
        'atm_pres_hpa': 1,
        'corr_atm_pres_red_mugal': 1,
        'norm_atm_pres_hpa': 1,
        'linear_scale': 5,
        'corr_oceanload_instrument_mugal': 1,
        'corr_oceanload_red_mugal': 1,
        'duration_sec': 0,
    }

    # Columns that will be shown in the table view, if available in the data (Also defines the order of columns):
    # - keys: Actual names of the dataframe columns
    # - items: Header names for the Table View Widget
    _SHOW_COLUMNS_IN_TABLE_DICT = {
        'keep_obs': 'Keep obs.',
        'station_name': 'Station',
        'setup_id': 'Setup ID',
        'loop_id': 'Loop ID',
        'lon_deg': 'Lon [°]',
        'lat_deg': 'Lat [°]',
        'alt_m': 'h [m]',
        'obs_epoch': 'Epoch UTC',
        'g_obs_mugal': 'g_obs [µGal]',
        'sd_g_obs_mugal': 'SD_obs [µGal]',
        'g_red_mugal': 'g_red [µGal]',
        'sd_g_red_mugal': 'SD_red [µGal]',
        'corr_terrain': 'Terrain corr.',
        'corr_tide_mugal': 'Instr. tide corr. [µGal]',
        'corr_tide_red_mugal': 'Tide corr. [µGal]',
        'temp': 'Temp. corr. [?]',
        'tiltx': 'Tilt_x [asec]',
        'tilty': 'Tilt_y [asec]',
        'dhb_m': 'dhb [m]',
        'dhf_m': 'dhf [m]',
        'vg_mugal': 'VG [µGal]',
        'duration_sec': 'Duration [sec]',
        'atm_pres_hpa': 'p [hPa]',
        'norm_atm_pres_hpa': 'pn [hPa]',
        'corr_atm_pres_red_mugal': 'p corr [µGal]',
        'linear_scale': 'lin. scale',
        'corr_oceanload_instrument_mugal': 'Instr. oceanload corr. [µGal]',
        'corr_oceanload_red_mugal': 'Oceanload corr. [µGal]',
    }
    _SHOW_COLUMNS_IN_TABLE = list(_SHOW_COLUMNS_IN_TABLE_DICT.keys())  # Actual list of columns to be shown

    # List of columns that are shown in the simple mode of the GUI
    _SHOW_COLUMNS_IN_TABLE_SIMPLE_GUI = [
        'station_name',
        'setup_id',
        'obs_epoch',
        'g_obs_mugal',
        'sd_g_obs_mugal',
        'g_red_mugal',
        'sd_g_red_mugal',
        'dhb_m',
        'dhf_m',
        'tiltx',
        'tilty',
        'corr_tide_red_mugal',
        'corr_tide_mugal',
        'vg_mugal',
        'duration_sec',
        'keep_obs',
        'atm_pres_hpa',
        'corr_atm_pres_red_mugal',
        'corr_oceanload_instrument_mugal',
        'corr_oceanload_red_mugal',
    ]

    def __init__(self, surveys):
        """Initialize the observation table view model.

        Parameters
        ----------
        surveys : dict of py.obj:`gravtools.survey.Survey` objects
        """
        self._surveys = {}
        self._data = None  # Observations (or at subset of them) of the survey with the name `self._data_survey_name`
        self._data_column_names = None
        QAbstractTableModel.__init__(self)
        self.load_surveys(surveys)
        self._data_survey_name = ''  # Name of the Survey that is currently represented by `self._data`
        self._setup_data = None  # Setup data (or at subset) of the survey with the name `self._data_survey_name`
        self.flag_gui_simple_mode = False
        # self._keep_obs_check_states = dict()  # To keep track of the checkbox states in the keep_obs column

    def reset_model(self):
        """Reset the model to a state without selected data to display."""
        self._data = None  # Observations (or at subset of them) of the survey with the name `self._data_survey_name`
        self._data_column_names = None
        self._data_survey_name = ''  # Name of the Survey that is currently represented by `self._data`
        self._setup_data = None  # Setup data (or at subset) of the survey with the name `self._data_survey_name`

    def load_surveys(self, surveys):
        """Load observation data (dict of survey objects in the campaign object) to the observation model.

        Notes
        -----
        The data is assigned by reference, i.e. all changes in `_surveys` will propagate to the data origin.
        """
        if surveys is None:
            self._surveys = {}
            return
        self._surveys = surveys

    def update_view_model(self, survey_name, setup_id, gui_simple_mode=False):
        """Update the `_data` data frame that hold the actual data that is viewed."""
        self.flag_gui_simple_mode = gui_simple_mode
        if survey_name is not None:
            try:
                obs_df = self._surveys[survey_name].obs_df  # Select the survey
                setup_df = self._surveys[survey_name].setup_df
            except KeyError:
                QMessageBox.critical(self.parent(), 'Error!', f'Survey "{survey_name}" is not available in this campaign.')
            else:
                self._data_survey_name = survey_name

                # Get list of columns to be depicted via the table model:
                # - Keep order of items in `self._SHOW_COLUMNS_IN_TABLE`
                obs_df_columns_set = frozenset(obs_df.columns)
                table_model_columns = [x for x in self.get_table_columns if x in obs_df_columns_set]

                if setup_id is None:  # No setup ID provided => Take all observations in survey
                    # self._data = obs_df.copy(deep=True)
                    self._data = obs_df.loc[:, table_model_columns].copy(deep=True)
                    if setup_df is None:
                        self._setup_data = None
                    else:
                        self._setup_data = setup_df.copy(deep=True)
                else:  # Only take observations of the specified setup
                    # self._data = obs_df[obs_df['setup_id'] == setup_id].copy(deep=True)
                    self._data = obs_df.loc[obs_df['setup_id'] == setup_id, table_model_columns].copy(deep=True)
                    if setup_df is None:
                        self._setup_data = None
                    else:
                        self._setup_data = setup_df[setup_df['setup_id'] == setup_id].copy(deep=True)
                # Column names of the actual table model:
                self._data_column_names = self._data.columns.to_list()
        else:
            self.reset_model()

    def headerData(self, section, orientation, role):
        # section is the index of the column/row.
        if self._data is not None:  # No data available yet
            if role == Qt.DisplayRole:
                if orientation == Qt.Horizontal:
                    return self._SHOW_COLUMNS_IN_TABLE_DICT[str(self._data.columns[section])]
                if orientation == Qt.Vertical:
                    return str(self._data.index[section])

    def rowCount(self, parent=None):
        if self._data is not None:
            return self._data.shape[0]
        else:
            return 0

    def columnCount(self, parent=None):
        if self._data is not None:
            return self._data.shape[1]
        else:
            return 0

    def data(self, index, role=Qt.DisplayRole):
        if self._data is not None:
            if index.isValid():
                if role == Qt.DisplayRole or role == Qt.UserRole:
                    value = self._data.iloc[index.row(), index.column()]

                    # Only checkboxes in the "keep_obs" flag column:
                    if role == Qt.DisplayRole:
                        if index.column() == self._data_column_names.index('keep_obs'):
                            return ''

                    # Custom formatter (string is expected as return type):
                    if value is None:  #
                        return NONE_REPRESENTATION_IN_TABLE_VIEW
                    elif isinstance(value, float):
                        if value != value:  # True, if value is "NaN"
                            return NONE_REPRESENTATION_IN_TABLE_VIEW
                        else:
                            try:
                                col_name_str = self._data_column_names[index.column()]
                                num_dec_places = self._DECIMAL_PLACES_PER_FLOAT_COLUMN[col_name_str]
                                return '{1:.{0}f}'.format(num_dec_places, value)
                            except Exception:
                                return str(value)
                    elif isinstance(value, np.int64):
                        return str(value)
                    elif isinstance(value, dt.datetime):
                        return value.strftime("%Y-%m-%d, %H:%M:%S")
                    else:  # all other
                        return str(value)

                if role == Qt.TextAlignmentRole:
                    value = self._data.iloc[index.row(), index.column()]
                    if isinstance(value, float) or isinstance(value, int) or isinstance(value, np.int64):
                        # Align right, vertical middle.
                        return Qt.AlignVCenter + Qt.AlignRight

                if role == Qt.BackgroundRole:
                    try:
                        if not self._data.iloc[index.row(), self._data_column_names.index('keep_obs')]:
                            return QtGui.QColor('red')
                    except Exception:
                        pass

                if role == Qt.CheckStateRole:
                    try:
                        if index.column() == self._data_column_names.index('keep_obs'):
                            keep_obs_flag = self._data.iloc[index.row(), self._data_column_names.index('keep_obs')]
                            if keep_obs_flag:
                                return Qt.Checked
                            else:
                                return Qt.Unchecked
                    except Exception:
                        print(f'ERROR: row: {index.row()}, col: {index.column()}')

    def flags(self, index):
        """Enable editing of table items."""
        flags = super(self.__class__, self).flags(index)
        flags |= Qt.ItemIsSelectable
        flags |= Qt.ItemIsEnabled
        if index.column() == self._data_column_names.index('keep_obs'):
            # flags |= Qt.ItemIsEditable  # No longer needed => Only checkbox in "keep_obs" column!
            flags |= Qt.ItemIsUserCheckable
        return flags

    def setData(self, index, value, role):
        """Example: https://www.semicolonworld.com/question/58510/how-to-display-a-pandas-data-frame-with-pyqt5"""
        if index.isValid():
            if role == Qt.EditRole:
                # Get column and row indices for dataframe:
                row = self._data.index[index.row()]
                col = self._data.columns[index.column()]
                if col == 'keep_obs':
                    # convert "value" (str) to bool and set item in dataframe:
                    if value == 'True':
                        self._data.at[row, col] = True
                        self.dataChanged.emit(index, index)  # Is it necessary?
                        # # Change data in `obs_df`:
                        # obs_df_row_index_int = self._data.index[index.row()]
                        # self._surveys[self._data_survey_name].obs_df.iat[
                        #     obs_df_row_index_int, Survey.get_obs_df_column_index('keep_obs')] = True
                    elif value == 'False':
                        self._data.at[row, col] = False
                        self.dataChanged.emit(index, index)  # Is it necessary?
                        # # Change data in `obs_df`:
                        # obs_df_row_index_int = self._data.index[index.row()]
                        # self._surveys[self._data_survey_name].obs_df.iat[
                        #     obs_df_row_index_int, Survey.get_obs_df_column_index('keep_obs')] = False
                    else:
                        QMessageBox.warning(self.parent(), 'Warning!',
                                            f'Input "{value}" not valid! Only "True" or "False" allowed.')
                        return False
                    return True  # Data successfully set

            if role == Qt.CheckStateRole:
                row = self._data.index[index.row()]
                col = self._data.columns[index.column()]
                idx_min = self.index(index.row(), 0)
                idx_max = self.index(index.row(), len(self._data_column_names) - 1)
                if col == 'keep_obs':
                    if value == Qt.Unchecked:
                        # print(f'row {index.row()}: Unchecked!')
                        self._data.at[row, col] = False
                        # Change data in `obs_df`:
                        self._surveys[self._data_survey_name].obs_df.iloc[
                            self._data.index[index.row()], Survey.get_obs_df_column_index('keep_obs')] = False
                        self.dataChanged.emit(idx_min, idx_max)  # Change color of obs. table row
                        self.dataChanged.emit(index, index, [9999])  # Change obs. plot and tree
                    elif value == Qt.Checked:
                        # print(f'row {index.row()}: Checked!')
                        self._data.at[row, col] = True
                        # Change data in `obs_df`:
                        self._surveys[self._data_survey_name].obs_df.iloc[
                            self._data.index[index.row()], Survey.get_obs_df_column_index('keep_obs')] = True
                        self.dataChanged.emit(idx_min, idx_max)  # Change color of obs. table row
                        self.dataChanged.emit(index, index, [9999])  # Change obs. plot and tree
                    else:
                        QMessageBox.warning(self.parent(), 'Warning!',
                                            f'Invalid value fpr keep observation flag: "{value}"')
                        return False

                return True  # Data successfully set

        return False

    @property
    def data_survey_name(self):
        return self._data_survey_name

    @property
    def get_data(self):
        return self._data

    @property
    def get_setup_data(self):
        return self._setup_data

    @property
    def get_table_columns(self):
        """Returns a list with all columns names of the dataframe that should be copied to the view model."""
        if self.flag_gui_simple_mode:
            return [value for value in self._SHOW_COLUMNS_IN_TABLE if value in self._SHOW_COLUMNS_IN_TABLE_SIMPLE_GUI]
        else:
            return self._SHOW_COLUMNS_IN_TABLE

    def get_data_df(self):
        """Returns the model data as pandas DataFrame."""
        print('test')



"""Dialog for loading CG6 observation files.

Copyright (C) 2024  Andreas Hellerschmied <andreas.hellerschmied@bev.gv.at>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
from PyQt5.QtWidgets import QDialog, QFileDialog, QMessageBox
from PyQt5.QtCore import Qt


from gravtools.gui.dialog_load_cg6_obs_files import Ui_DialogLoadCg6ObservationFiles
from gravtools import settings


class DialogLoadCg6ObservationFiles(QDialog, Ui_DialogLoadCg6ObservationFiles):
    """Dialog for loading CG6 observation files in different formats."""

    def __init__(self, parent=None):
        super().__init__(parent)

        # Run the .setupUi() method to show the GUI
        self.setupUi(self)

        # Fill comboboxes:
        for key, item in settings.CG6_SURVEY_DATA_SOURCE_TYPES_SHORT.items():
            self.comboBox_format_selection.addItem(item, userData=key)
            idx = self.comboBox_format_selection.findData(key)
            tooltip = settings.CG6_SURVEY_DATA_SOURCE_TYPES[key]
            self.comboBox_format_selection.setItemData(idx, tooltip, Qt.ToolTipRole)

        for key, item in settings.CG6_OBS_FILE_LOCATION_TYPES.items():
            self.comboBox_location_type.addItem(item['label'], userData=key)
            idx = self.comboBox_location_type.findData(key)
            self.comboBox_location_type.setItemData(idx, item['tooltip'], Qt.ToolTipRole)

        for key, item in settings.CG6_OBS_FILE_ERROR_TYPES.items():
            self.comboBox_uncertainty_type.addItem(item['label'], userData=key)
            idx = self.comboBox_uncertainty_type.findData(key)
            self.comboBox_uncertainty_type.setItemData(idx, item['tooltip'], Qt.ToolTipRole)

        for key, item in settings.CG6_PRESSURE_IN_COLUMN.items():
            self.comboBox_pres_in_column.addItem(item['label'], userData=key)
            idx = self.comboBox_pres_in_column.findData(key)
            self.comboBox_pres_in_column.setItemData(idx, item['tooltip'], Qt.ToolTipRole)

        for key, item in settings.CG6_DHB_IN_COLUMN.items():
            self.comboBox_dhb_in_column.addItem(item['label'], userData=key)
            idx = self.comboBox_dhb_in_column.findData(key)
            self.comboBox_dhb_in_column.setItemData(idx, item['tooltip'], Qt.ToolTipRole)

        # Connect signals and slots:
        self.comboBox_format_selection.currentIndexChanged.connect(self.on_combobox_current_index_changed)
        self.pushButton_select_files.pressed.connect(self.select_files)
        self.pushButton_clear_list.pressed.connect(self.listWidget_open_files.clear)
        self.comboBox_pres_in_column.currentIndexChanged.connect(self.check_alternative_column_data_selection)
        self.comboBox_dhb_in_column.currentIndexChanged.connect(self.check_alternative_column_data_selection)

        # Set default values for GUI items (AFTER connecting signals and slots):
        idx = self.comboBox_format_selection.findData(settings.CG6_SURVEY_DATA_SOURCE_TYPE_DEFAULT)
        self.comboBox_format_selection.setCurrentIndex(idx)
        idx = self.comboBox_pres_in_column.findData(settings.CG6_PRESSURE_IN_COLUMN_DEFAULT)
        self.comboBox_pres_in_column.setCurrentIndex(idx)
        idx = self.comboBox_dhb_in_column.findData(settings.CG6_DHB_IN_COLUMN_DEFAULT)
        self.comboBox_dhb_in_column.setCurrentIndex(idx)
        idx = self.comboBox_uncertainty_type.findData(settings.CG6_OBS_FILE_ERROR_TYPES_DEFAULT)
        self.comboBox_uncertainty_type.setCurrentIndex(idx)
        idx = self.comboBox_location_type.findData(settings.CG6_OBS_FILE_LOCATION_TYPES_DEFAULT)
        self.comboBox_location_type.setCurrentIndex(idx)

    def check_alternative_column_data_selection(self):
        """Check selection in comboboxes for alternative data in columns and raise a warning."""
        dhb_data = self.comboBox_dhb_in_column.currentData()
        pres_data = self.comboBox_pres_in_column.currentData()
        if (dhb_data == pres_data) and not (dhb_data == 'no_data' and pres_data == 'no_data'):
            QMessageBox.warning(self, 'Warning!', f'Non-unique assignment of alternative data in observation file '
                                                  f'columns: {settings.CG6_DHB_IN_COLUMN[dhb_data]["label"]}.\n\n'
                                                  f'Please change the column selection!')

    def on_combobox_current_index_changed(self, idx):
        """On combobox index change."""
        if idx == -1:  # Invalid index
            return
        format_key = self.comboBox_format_selection.itemData(idx)
        # Update format description label
        description = settings.CG6_SURVEY_DATA_SOURCE_TYPES[format_key]
        self.label_format_description.setText(description)

    def select_files(self):
        """Select observation files using a file selection dialog"""
        options = QFileDialog.Options()
        options |= QFileDialog.DontUseNativeDialog
        obs_file_filenames, _ = QFileDialog.getOpenFileNames(self,
                                                             'Select CG6 observation file(s)',
                                                             self.parent().campaign.output_directory,
                                                             "CG6 observation file (*.DAT *.TXT)",
                                                             options=options)
        if not obs_file_filenames:
            return
        self.listWidget_open_files.clear()
        self.listWidget_open_files.addItems(obs_file_filenames)

    @property
    def file_list(self) -> list:
        """Return a list of all files in the list widget."""
        # self.listWidget_open_file.
        return [self.listWidget_open_files.item(idx).text() for idx in range(self.listWidget_open_files.count())]

    @property
    def file_format(self) -> str:
        """Returns the file format identifier string."""
        return self.comboBox_format_selection.currentData()

    @property
    def location_type(self) -> str:
        """Returns the location type identifier."""
        return self.comboBox_location_type.currentData()

    @property
    def error_type(self) -> str:
        """Returns the error type identifier."""
        return self.comboBox_uncertainty_type.currentData()

    @property
    def pres_in_column(self) -> str:
        """Returns the column identifier for obs_df for pressure data."""
        return self.comboBox_pres_in_column.currentData()

    @property
    def dhb_in_column(self) -> str:
        """Returns the column identifier for height differences dhb."""
        return self.comboBox_dhb_in_column.currentData()



"""Dialog for managing time series data for observation corrections.

Copyright (C) 2023  Andreas Hellerschmied <andreas.hellerschmied@bev.gv.at>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import os
import pytz
import datetime as dt

from PyQt5.QtWidgets import QDialog, QMessageBox, QTreeWidgetItem
import pyqtgraph as pg

from gravtools.gui.dialog_correction_time_series import Ui_DialogCorrectionTimeSeries
from gravtools.gui.dlg_load_tsf_file import DialogLoadTsfFile
from gravtools import __version__
from gravtools import settings

# Set splitter stretch:
_SPLITTER_STRETCH_LEFT = 1
_SPLITTER_STRETCH_RIGHT = 10


class TimeAxisItem(pg.AxisItem):
    """"Needed to handle the x-axes tags representing date and time.
    From: https://stackoverflow.com/questions/49046931/how-can-i-use-dateaxisitem-of-pyqtgraph

    Notes
    -----
    The timestamps need to refer to UTC!"
    """
    def tickStrings(self, values, scale, spacing) -> str:
        """Handles the x-axes tags representing date and time."""
        return [dt.datetime.fromtimestamp(value, tz=pytz.utc).strftime(settings.Y_TICK_DATETIME_FORMAT) for value in values]


class DialogCorrectionTimeSeries(QDialog, Ui_DialogCorrectionTimeSeries):
    """Dialog for managing time series data for observation corrections."""

    def __init__(self, parent=None):
        super().__init__(parent)

        # Run the .setupUi() method to show the GUI
        self.setupUi(self)

        # Set the relative size of the left and right layout of the splitter:
        self.splitter.setStretchFactor(_SPLITTER_STRETCH_LEFT, _SPLITTER_STRETCH_RIGHT)

        # Connect signals and slots:
        self.pushButton_load_tsf_file.pressed.connect(self.launch_dlg_load_tsf_file)
        self.pushButton_collapse.pressed.connect(self.collapse_all_tree_widget)
        self.pushButton_expand.pressed.connect(self.expand_all_tree_widget)
        self.treeWidget_selection.currentItemChanged.connect(self.tree_widget_current_item_changed)
        self.pushButton_delete_selection.pressed.connect(self.delete_selected_item)

        # Attributes
        self.dlg_load_tsf_file = DialogLoadTsfFile(self)

        # Other initial settings:
        self.reset_stacked_widget()
        self.set_up_time_series_plot_widget()

    def launch_dlg_load_tsf_file(self):
        """Launch dialog to load TSF file."""
        if self.parent().campaign is None:
            QMessageBox.critical(self, 'Error!', 'No Campaign available! Please create a campaign first.')
            return

        self.check_correction_time_series_object()

        # Launch file selection dialog:
        return_value = self.dlg_load_tsf_file.exec()
        if return_value != QDialog.Accepted:
            return  # Do not load any data

        # Get filename and other parameters from the file selection dialog:
        filename_tsf = self.dlg_load_tsf_file.lineEdit_filename.text()
        survey_name = self.dlg_load_tsf_file.lineEdit_survey_name.text()
        filter_location = self.dlg_load_tsf_file.lineEdit_filter_location.text()
        filter_instrument = self.dlg_load_tsf_file.lineEdit_filter_instrument.text()
        filter_data_type = self.dlg_load_tsf_file.lineEdit_filter_data_type.text()
        overwrite_channel = self.dlg_load_tsf_file.checkBox_overwrite_channel.isChecked()

        is_effect = self.dlg_load_tsf_file.radioButton_effect.isChecked()
        is_correction = self.dlg_load_tsf_file.radioButton_correction.isChecked()

        if not os.path.exists(filename_tsf):
            QMessageBox.critical(self, 'Error!',
                                 f"The selected TSF file ({filename_tsf}) does not exist!")
            return

        # Load the TSF file:
        try:
            self.parent().campaign.correction_time_series.load_tfs_file(survey_name=survey_name,
                                                                        filename_tsf=filename_tsf,
                                                                        location=filter_location,
                                                                        instrument=filter_instrument,
                                                                        data_type=filter_data_type,
                                                                        overwrite_channel=overwrite_channel,
                                                                        is_correction=is_correction)
        except Exception as e:
            QMessageBox.critical(self, 'Error!', str(e))

        self.reset_update_gui()

    def check_correction_time_series_object(self, parent=None):
        """Adds a correction time series object to the campaign, if missing."""
        if not hasattr(self.parent().campaign, 'correction_time_series'):
            self.parent().campaign.add_empty_correction_time_series()
            if parent is None:
                parent = self
            QMessageBox.warning(parent, 'Warning!', f'The campaign "{self.parent().campaign.campaign_name}" had no '
                                                  f'"correction_time_series" attribute. This '
                                                  f'may happens when loading campaign data from previous GravTools '
                                                  f'versions. Current version: {__version__}. '
                                                  f'Campaign data version: {self.parent().campaign.gravtools_version}. '
                                                  f'However, "correction_time_series" was added to the campaign. Please'
                                                  f'save the campaign to keep the changes!')

    def reset_update_gui(self):
        """Update the GUI after changing data, e.g. loading from TSF files or deleting data."""
        self.reset_stacked_widget()
        self.populate_survey_tree_widget()
        self.update_timer_series_plot()

    def populate_survey_tree_widget(self):
        """Populate the survey tree widget."""

        correction_time_series = self.parent().campaign.correction_time_series

        # Delete existing items:
        self.treeWidget_selection.blockSignals(True)
        self.treeWidget_selection.clear()

        # Add new items:
        # Parent items (surveys):
        for survey_name, survey_corrections in correction_time_series.surveys.items():
            parent = QTreeWidgetItem(self.treeWidget_selection)
            parent.setText(0, survey_name)
            for station_name, station_correction in survey_corrections.stations.items():
                child = QTreeWidgetItem(parent)
                child.setText(0, station_name)

        self.treeWidget_selection.show()
        self.treeWidget_selection.blockSignals(False)
        self.treeWidget_selection.expandToDepth(0)  # Expand items to a certain depth

    def tree_widget_current_item_changed(self, tree_item, column):
        """Whenever the current item changes in the selection tree widget."""
        # Show item details in stacked widget according to selection:
        depth = get_depth(tree_item)
        if depth == 1:
            survey_name = tree_item.text(0)
            self.stacked_widget_show_survey_details(survey_name=survey_name)
        elif depth == 2:
            station_name = tree_item.text(0)
            survey_name = tree_item.parent().text(0)
            self.stacked_widget_show_station_details(survey_name=survey_name, station_name=station_name)
            self.update_timer_series_plot()
        else:
            pass

    def stacked_widget_show_station_details(self, survey_name, station_name):
        """Show station details in the stacked widget."""
        self.stackedWidget_item_details.setCurrentIndex(2)
        correction_time_series = self.parent().campaign.correction_time_series
        tidal_correction = correction_time_series.surveys[survey_name].stations[station_name].tidal_correction
        self.label_station_station_name.setText(station_name)
        self.label_station_data_source.setText(tidal_correction.data_source)
        self.label_station_description.setText(tidal_correction.description)
        self.label_station_unit.setText(tidal_correction.unit)
        self.label_station_created.setText(tidal_correction.created_datetime_utc_str)
        self.label_station_model_type.setText(tidal_correction.model_type)
        self.label_station_start.setText(tidal_correction.start_datetime_str)
        self.label_station_end.setText(tidal_correction.end_datetime_str)
        self.label_station_duration.setText(tidal_correction.duration_dhms_str)
        self.label_station_nuber_of_datapoints.setText(str(tidal_correction.number_of_datapoints))

    def stacked_widget_show_survey_details(self, survey_name):
        """Show survey details in the stacked widget."""

        correction_time_series = self.parent().campaign.correction_time_series
        self.stackedWidget_item_details.setCurrentIndex(1)
        self.label_survey_name.setText(survey_name)
        self.label_num_of_stations.setText(str(correction_time_series.surveys[survey_name].number_of_stations))

    def collapse_all_tree_widget(self):
        """Collapse all items in selection tree widget."""
        self.treeWidget_selection.collapseAll()

    def expand_all_tree_widget(self):
        """Expand all items in tree widget."""
        self.treeWidget_selection.expandAll()

    def reset_stacked_widget(self):
        """Reset the stacked widget by displaying page with index 0"""
        self.stackedWidget_item_details.setCurrentIndex(0)

    def delete_selected_item(self):
        """Delete the selected item in the tree widget"""
        correction_time_series = self.parent().campaign.correction_time_series
        tree_item = self.treeWidget_selection.currentItem()
        depth = get_depth(tree_item)
        if depth == 1:  # Survey
            survey_name = tree_item.text(0)
            msg_text = f'Survey {survey_name}'
            reply = QMessageBox.question(self,
                                         'Delete survey corrections?',
                                         msg_text,
                                         QMessageBox.Yes | QMessageBox.No, QMessageBox.No)
            if reply == QMessageBox.Yes:
                try:
                    correction_time_series.delete_survey_correction(survey_name)
                except KeyError:
                    QMessageBox.critical(self, 'Error!', f'No corrections for survey "{survey_name}" available!')
                except Exception as e:
                    QMessageBox.critical(self, 'Error!', str(e))
                else:
                    self.reset_update_gui()
            else:
                return
        elif depth == 2:  # Station
            station_name = tree_item.text(0)
            survey_name = tree_item.parent().text(0)
            msg_text = f'Station {survey_name}'
            reply = QMessageBox.question(self,
                                         'Delete station corrections?',
                                         msg_text,
                                         QMessageBox.Yes | QMessageBox.No, QMessageBox.No)
            if reply == QMessageBox.Yes:
                try:
                    correction_time_series.surveys[survey_name].delete_station_correction(station_name)
                except KeyError:
                    QMessageBox.critical(self, 'Error!', f'No corrections for station "{station_name}" in survey "{survey_name}" available!')
                except Exception as e:
                    QMessageBox.critical(self, 'Error!', str(e))
                else:
                    self.reset_update_gui()
            else:
                return
        else:
            return

    def set_up_time_series_plot_widget(self):
        """Set up `self.graphicsLayoutWidget_results_vg_plot` widget."""
        self.glw_ts_plot = self.graphicsLayoutWidget_time_series
        self.glw_ts_plot.setBackground('w')  # white background color
        self.ts_plot = self.glw_ts_plot.addPlot(0, 0, name='ts_plot',
                                                      axisItems={'bottom': TimeAxisItem(orientation='bottom')})
        # self.ts_plot.setLabel(axis='bottom', text='Time')  # Takes too much space
        self.ts_plot.addLegend()
        self.ts_plot.showGrid(x=True, y=True)
        # self.ts_plot.setTitle('')  # Takes too much space

    def plot_time_series(self, data, timestamps, plot_name: str ='', y_label: str = ''):
        # Clear plot in any case:
        self.ts_plot.clear()
        self.plot_xy_data(self.ts_plot, timestamps, data, plot_name=plot_name, color='b', symbol='o', symbol_size=8)
        self.ts_plot.setLabel(axis='left', text=y_label)
        self.ts_plot.autoRange()

    @staticmethod
    def plot_xy_data(plot_item, x, y, plot_name, color='k', symbol='o', symbol_size=10):
        """Plot XY-data."""
        pen = pg.mkPen(color=color)
        plot_item.plot(x, y, name=plot_name, pen=pen, symbol=symbol, symbolSize=symbol_size, symbolBrush=(color))

    def update_timer_series_plot(self):
        """Update the time series plots according to the selection in the tree widget."""
        tree_item = self.treeWidget_selection.currentItem()
        if tree_item is None:
            return
        depth = get_depth(tree_item)
        if depth == 2:  # Time series data
            # Get the data:
            station_name = tree_item.text(0)
            survey_name = tree_item.parent().text(0)
            try:
                time_series = self.parent().campaign.correction_time_series.surveys[survey_name].stations[
                    station_name].tidal_correction
            except KeyError:
                return
            if time_series is None:
                return
            data = time_series.data
            timestamps = time_series.ref_time_unix
            # Plot the data:
            try:
                self.plot_time_series(data, timestamps, plot_name=time_series.description, y_label=time_series.unit)
            except Exception as e:
                QMessageBox.critical(self, 'Error!', str(e))
        else:
            self.ts_plot.clear()
            return


def get_subtree_nodes(tree_widget_item) -> list:
    """Returns al QTreeWidgetItems in a subtree rooted at a given node (recursive)."""
    nodes = []
    for i in range(tree_widget_item.childCount()):
        nodes.extend(get_subtree_nodes(tree_widget_item.child(i)))
    return nodes


def get_all_items(tree_widget) -> list:
    """Returns all QTreeWidgetItems in a given QTreeWidget."""
    all_items = []
    for i in range(tree_widget.topLevelItemCount()):
        top_item = tree_widget.topLevelItem(i)
        all_items.extend(get_subtree_nodes(top_item))
    return all_items


def get_depth(tree_widget_item) -> int:
    """Returns the depth of a QTreeWidgetItem within the QTreeWidget."""
    depth = 0
    while tree_widget_item is not None:
        depth += 1
        tree_widget_item = tree_widget_item.parent()
    return depth








# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'gravtools/gui/dialog_load_stations.ui'
#
# Created by: PyQt5 UI code generator 5.15.9
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_Dialog_load_stations(object):
    def setupUi(self, Dialog_load_stations):
        Dialog_load_stations.setObjectName("Dialog_load_stations")
        Dialog_load_stations.resize(464, 119)
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(Dialog_load_stations)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.verticalLayout = QtWidgets.QVBoxLayout()
        self.verticalLayout.setSpacing(20)
        self.verticalLayout.setObjectName("verticalLayout")
        self.formLayout = QtWidgets.QFormLayout()
        self.formLayout.setObjectName("formLayout")
        self.pushButton_select_oesgn_table = QtWidgets.QPushButton(Dialog_load_stations)
        self.pushButton_select_oesgn_table.setObjectName("pushButton_select_oesgn_table")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.pushButton_select_oesgn_table)
        self.lineEdit_oesgn_table_file_path = QtWidgets.QLineEdit(Dialog_load_stations)
        self.lineEdit_oesgn_table_file_path.setObjectName("lineEdit_oesgn_table_file_path")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.lineEdit_oesgn_table_file_path)
        self.verticalLayout.addLayout(self.formLayout)
        self.buttonBox = QtWidgets.QDialogButtonBox(Dialog_load_stations)
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.verticalLayout.addWidget(self.buttonBox)
        self.verticalLayout_2.addLayout(self.verticalLayout)

        self.retranslateUi(Dialog_load_stations)
        self.buttonBox.accepted.connect(Dialog_load_stations.accept) # type: ignore
        self.buttonBox.rejected.connect(Dialog_load_stations.reject) # type: ignore
        QtCore.QMetaObject.connectSlotsByName(Dialog_load_stations)
        Dialog_load_stations.setTabOrder(self.pushButton_select_oesgn_table, self.lineEdit_oesgn_table_file_path)

    def retranslateUi(self, Dialog_load_stations):
        _translate = QtCore.QCoreApplication.translate
        Dialog_load_stations.setWindowTitle(_translate("Dialog_load_stations", "Load stations"))
        self.pushButton_select_oesgn_table.setToolTip(_translate("Dialog_load_stations", "Select ÖSGN table file"))
        self.pushButton_select_oesgn_table.setText(_translate("Dialog_load_stations", "Select ÖSGN Table"))
        self.lineEdit_oesgn_table_file_path.setToolTip(_translate("Dialog_load_stations", "Enter path to ÖSGN table file"))



"""
Utility functions for gravtools.
This module is part of the gravtools package.
Author: Andreas Hellerschmied
"""

# Imports from other gravtools modules:
from bev_legacy import settings


def corr_instrument_ref_heights(obs_df):
    """
    Korrektur der Höhenoffsets zwischen Instrument-Oberkante (bei Messung notiert) und Boden bzw. Festpunkt (dHF, dHB)
    um den Höhenoffset zwischen Instrumenten-Oberkante und Messwerk zu berücksichtigen. Diese Korrektur ist abhängig vom
    Intrumenten-Typ.
    :param obs_df: dataframe mit Messungen
    :return: dataframe mit Messungen (erweitert)
    """

    # ### Je nach Gerätetyp (CG5, CG3) die Bezugs-Höhen (dhb, dhf) anpassen ###:
    for grav_typ, corr_m in settings.GRAVIMETER_REFERENCE_HEIGHT_CORRECTIONS_m.items():
        obs_df.loc[obs_df['gravimeter_typ'] == grav_typ, 'dhb_m'] = \
            obs_df.loc[obs_df['gravimeter_typ'] == grav_typ, 'dhb_m'] + corr_m
        obs_df.loc[obs_df['gravimeter_typ'] == grav_typ, 'dhf_m'] = \
            obs_df.loc[obs_df['gravimeter_typ'] == grav_typ, 'dhf_m'] + corr_m
    return obs_df


def red_g_obs_vg(obs_df):
    """
    Gravimeter-Lesungen mittels Vertikalgadienten auf die Höhe des Festpunktes reduzieren (mittels dHF)
    :param obs_df: dataframe mit Messungen
    :return: obs_df: dataframe mit Messungen (mit auf die Festpunkte reduzierten Schweremessungen)
    """
    obs_df['g_red_mugal'] = obs_df['g_mugal'] + obs_df['vg_mugal'] * obs_df['dhf_m']
    return obs_df


def prep_polyval_coef(pol_coef_dict):
    """
    Vorbereitung der Polynomkoeffizienten für np.polyval().
    :param pol_coef_dict: Dictionary with polynomial coefficients from mlr.
    :return: poly_coef: list of polynomial coeficients suitable for np.polyval()
    """
    poly_coef = list()
    for degree, value in pol_coef_dict.items():
        poly_coef.append(value)
    poly_coef.reverse()
    poly_coef.append(0)
    return poly_coef


"""Classes for least-squares adjustment of gravimeter-surveys.

Copyright (C) 2021  Andreas Hellerschmied <andreas.hellerschmied@bev.gv.at>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import numpy as np
from scipy import stats
import datetime as dt
import pytz
import copy
import pandas as pd

from gravtools.models.misc import time_it, get_nonunique_items

# optional imports:
try:
    import geopandas
except ImportError:
    _has_geopandas = False
else:
    _has_geopandas = True

from gravtools import settings


class LSM:
    """Base class for LSM classes.

    Attributes
    ----------
    lsm_method : str
        Defines the adjustment method. Has to b listed in :py:obj:`gravtools.settings.ADJUSTMENT_METHODS`.
    stat_df : :py:obj:`gravtools.Station.stat_df`
        The station dataframe contains all relevant station data.
    setups : dict of dicts
        The setups dictionary contains all observation data used for the adjustment. The keys of the dictionary
        are the survey names (str). The items are again keys with the following items:

        - ref_epoch_delta_t_h : datetime object
            Reference epoch for the relative reference times in the column `delta_t_h` in the `setup_df` dataframe.
            The reference epoch is determined as the epoch of the first (active) observation in this survey.
        -  ref_epoch_delta_t_campaign_h : datetime object
            Reference epoch for the relative reference times in the column `delta_t_campaign_h` in the `setup_df`
            dataframe. The reference epoch is determined as the epoch of the first (active) observation in the campaign.
        - setup_df : Pandas DataFrame
            Pandas dataframes containing the observation data (see :py:obj:`gravtools.Survey.setup_df`).
        - tide_correction_type : str, optional (default='')
            Type of the tidal corrections applied on the reduced observations (column `g_red_mugal` in `obs_df`) that
            were used to calculate the setup data store in this dict. Valid entries have to be listed in
            :py:obj:`gravtools.settings.TIDE_CORRECTION_TYPES`.
        - reference_height_type : str, optional (default='')
            Reference height type of the reduced observations (column `g_red_mugal` in `obs_df`) that
            were used to calculate the setup data store in this dict. Valid entries have to be listed in
            :py:obj:`gravtools.settings.REFERENCE_HEIGHT_TYPE`.

    comment : str, optional (default = '')
        Optional comment on the adjustment run.
    init_time : datetime object
        Representing the date and time the LSM object has been initialized.
    write_log : bool, optional (default=True)
        Flag that indicates whether log string should be written or not.
    log_str : str
        String to log the status of the adjustment.
    setup_obs_df : Pandas DataFrame
        Pandas Dataframes for logging (differential or absolute) setup observations, the related metadata, estimation
        results and statistics. The columns of the dataframe may differ between adjustment methods.
    number_of_iterations : int (default=0)
        Indicates the number of iterations if iterative adjustment was applied. `0` indicated that a non-iterative
        adjustment was applied.
    drift_ref_epoch_type : string ('survey' or 'campaign'), optional (default='survey')
        Defines whether the reference epoch t0 for the estimation of the drift polynomials for each survey in the
        campaign is the reference epoch of the first (active) observation in each survey (option: 'survey') or the
        first (active) observation in the whole campaign (option: 'campaign').
        """

    def __init__(self, lsm_method, stat_df, setups, comment='', write_log=True):
        """
        Parameters
        ----------
        lsm_method : str
            Defines the adjustment method. Has to b listed in :py:obj:`gravtools.settings.ADJUSTMENT_METHODS`.
        stat_df : :py:obj:`gravtools.Station.stat_df`
            The station dataframe contains all relevant station data.
        setups : dict of dicts
            The setups dictionary contains all observation data used for the adjustment. The keys of the dictionary
            are the survey names (str). The items are again keys with the following items:

            - ref_epoch_delta_t_h : datetime object
                Reference epoch for the relative reference times in the column `delta_t_h` in the `setup_df` dataframe.
                The reference epoch is determined as the epoch of the first (active) observation in this survey.
            -  ref_epoch_delta_t_campaign_h : datetime object
                Reference epoch for the relative reference times in the column `delta_t_campaign_h` in the `setup_df`
                dataframe. The reference epoch is determined as the epoch of the first (active) observation in the campaign.
            - setup_df : Pandas DataFrame
                Pandas dataframes containing the observation data (see :py:obj:`gravtools.Survey.setup_df`).
            - tide_correction_type : str, optional (default='')
                Type of the tidal corrections applied on the reduced observations (column `g_red_mugal` in `obs_df`) that
                were used to calculate the setup data store in this dict. Valid entries have to be listed in
                :py:obj:`gravtools.settings.TIDE_CORRECTION_TYPES`.
        - red_reference_height_type : str, optional (default='')
                Reference height type of the reduced observations (column `g_red_mugal` in `obs_df`) that
                were used to calculate the setup data store in this dict. Valid entries have to be listed in
                :py:obj:`gravtools.settings.REFERENCE_HEIGHT_TYPE`.

        comment : str, optional (default = '')
            Optional comment on the adjustment run.
        write_log : bool, optional (default=True)
            Flag that indicates whether log string should be written or not.

        Notes
        -----
        In order to prevent altering the original input data (in case of assignment via reference), deep copies of the
        input data are stored in objects of this class.
        """
        # Initial input checks:
        if lsm_method in settings.ADJUSTMENT_METHODS.keys():
            self.lsm_method = lsm_method
        else:
            raise ValueError(f'"{lsm_method}" is not a valid adjustment method identifier! All valid methods are '
                             f'defined in gravtools.settings.ADJUSTMENT_METHODS.')
        if isinstance(comment, str):
            self.comment = comment
        else:
            raise ValueError(f'"{comment}" need to be a string.')
        if isinstance(write_log, bool):
            self.write_log = write_log
        else:
            raise ValueError(f'"{write_log}" needs to be a boolean.')
        self.init_time = dt.datetime.now(tz=pytz.utc)
        self.log_str = ''

        # Create deep copies of the input data items:
        self.stat_df = stat_df.copy(deep=True)
        self.setups = copy.deepcopy(setups)

        # Initialize attributes:
        self.setup_obs_df = None  # Dataframe that contain the observation (setup) related results
        self.observed_stations = None  # Unique list of observed stations; defines the station IDs for matrices
        self.stat_obs_df = None  # Station dataframe that contains estimation results of observed stations
        self.drift_pol_df = None  # DataFrame that contains the estimated parameters of the drift polynomials and their statistics for each survey
        self.vg_pol_df = None  # DataFrame that contains the estimated parameters of the vertical gravity gradient polynomial

        # Estimation settings:
        self.drift_polynomial_degree = None
        self.sig0_a_priori = None  # A priori standard deviation of unit weight
        self.scaling_factor_datum_observations = None
        self.confidence_level_chi_test = None
        self.confidence_level_tau_test = None
        self.drift_ref_epoch_type = ''  # 'survey' or 'campaign'
        self.vg_polynomial_degree = None
        self.vg_polynomial_ref_height_offset_m = 0.0

        # General statistics:
        self.number_of_stations = None
        self.number_of_datum_stations = None
        self.number_of_estimates = None
        self.degree_of_freedom = None

        # A posteriori statistics:
        self.s02_a_posteriori = None  # A posteriori variance of unit weight

        # Matrices:
        self.Cxx = None  # Co-variance matrix of estimated parameters
        self.Rxx = None  # Correlation matrix of estimates parameters
        self.x_estimate_names = []  # List of items in the Cxx matrix (index 0 to n)
        self.mat_A = None  # Design matrix
        self.mat_x = None  # Estimated parameters

        # Iterative adjustment:
        self.number_of_iterations = 0  # `0` indicates no iterations.

        # Statistical tests:
        self.number_of_outliers = None
        self.global_model_test_status = ''  # str

    @classmethod
    def from_campaign(cls, campaign, comment='', write_log=True):
        """Constructor that generates and populates the LSM object (child!) from a Campaign class object.

        Notes
        -----
        Put all checks and data preparations relevant for all LSM methods into this method!

        Parameters
        ----------
        campaign : :py:obj:`gravtools.models.survey.Campaign`
            The campaign object needs to provide setup data for all active surveys and the related station data.
        comment : str, optional (default = '')
            Arbitrary comment on the LSM run.
        write_log : bool, optional (default=True)
            Flag that indicates whether log string should be written or not.

        Returns
        -------
        LSM Object of the child class
            Contains all information required for adjusting the campaign.
        """
        # Check if all required data is available in the campaign object:

        # Comment:
        if not isinstance(comment, str):
            raise TypeError(f'"comment" needs to be a string!')

        # Station data:
        if campaign.stations is None:
            raise AssertionError(f'The campaign "{campaign.campaign_name}" does not contain any station data!')
        else:
            if not hasattr(campaign.stations, 'stat_df'):
                raise AssertionError(f'The campaign "{campaign.campaign_name}" has not station dataframe!')
            else:
                if len(campaign.stations.stat_df) == 0:
                    raise AssertionError(f'The campaign "{campaign.campaign_name}" has an empty station dataframe!')

        # Survey data:
        if campaign.surveys is None:
            raise AssertionError(f'The campaign "{campaign.campaign_name}" does not contain any survey data!')
        else:
            if len(campaign.surveys) == 0:
                raise AssertionError(f'The campaign "{campaign.campaign_name}" contains no survey data!')

        # Create setups dict:
        # Loop over surveys in campaign:
        setups = {}
        for survey_name, survey in campaign.surveys.items():
            if survey.keep_survey and survey.is_active:
                if survey.setup_df is None:
                    raise AssertionError(f'Setup data is missing for survey "{survey_name}"')
                else:
                    if len(survey.setup_df) < 2:
                        raise AssertionError(f'Survey "{survey_name}" has less than two setup observations! '
                                             f'A minimum of two setups is required in order to define '
                                             f'differential observations!')
                    else:
                        setup_data_dict = {'ref_epoch_delta_t_h': survey.ref_delta_t_dt,
                                           'ref_epoch_delta_t_campaign_h': campaign.ref_delta_t_dt,
                                           'setup_df': survey.setup_df,
                                           'tide_correction_type': survey.setup_tide_correction_type,
                                           'reference_height_type': survey.setup_reference_height_type,
                                           'atm_pres_correction_type': survey.setup_atm_pres_correction_type,
                                           'scale_correction_type': survey.setup_scale_correction_type,
                                           'setup_calc_method': survey.setup_calc_method,
                                           'setup_obs_list_df': survey.setup_obs_list_df}
                        setups[survey_name] = setup_data_dict

        # Check if setup data is available:
        if len(setups) == 0:
            raise AssertionError(f'Setup data of campaign "{campaign.campaign_name}" does not contain observations!')

        # Initialize and return LSM object:
        return cls(campaign.stations.stat_df, setups, comment=comment, write_log=write_log)

    def adjust_autoscale_s0(self,
                            iteration_approach='Multiplicative',
                            s02_target=1,
                            s02_target_delta=0.1,
                            max_number_iterations=10,
                            add_const_to_sd_of_observations_step_size_mugal=5.0,
                            max_total_additive_const_to_sd_mugal=20.0,
                            multiplicative_factor_step_size_percent=10.0,  # [%]
                            max_multiplicative_factor_to_sd_percent=200.0,  # [%]
                            min_multiplicative_factor_to_sd_percent=50.0,  # [%]
                            drift_pol_degree=1,
                            sig0_mugal=1,
                            scaling_factor_datum_observations=1.0,
                            add_const_to_sd_of_observations_mugal=0.0,
                            scaling_factor_for_sd_of_observations=1.0,
                            confidence_level_chi_test=0.95,
                            confidence_level_tau_test=0.95,
                            drift_ref_epoch_type='survey',
                            noise_floor_mugal=0.0,
                            verbose=False,
                            ):  # Confidence level):
        """Run the adjustment iteratively in order to adjust s0 to the target value by adapting the SD of observations.

        Notes
        -----
        Iterative adjustment works for differential and non-differential LSM adjustment, not for VG estimation!

        Parameters
        ---------
        iteration_approach : str, optional (default='Multiplicative')
            Defines the iteration approach (Multiplicative od Additive)
        s02_target : float, optional (default=1.0)
            Target a posteriori s0² (variance of unit weight) for the iteration.
        s02_target_delta : float, optional (default=0.1)
            Permissible deviation of the target a posteriori s0². As soon as the a posteriori s0 of the current lsm run
            lies withing the threshold of `s02_target` +- `s02_target_delta` the iteration was successful and stops.
        max_number_iterations : int, optional (default=10)
            Maximum allowed number of iterations. If the a posteriori s0 does not lie within the defined threshold
            (`s02_target` +- `s02_target_delta`) after `max_number_iterations` iterations an assertion error is raised.
        add_const_to_sd_of_observations_step_size_mugal : float, optional (default=5.0)
            Initial iteration step size for the additive constant that is added to the standard deviation of all
            observations. This parameter is only considered when using the `additive` iteration approach.
        max_total_additive_const_to_sd_mugal : float, optional (default=20.0)
            If the additive factor for the SD of observations that is determined iteratively in order to reach the
            defined target s0 is larger than `max_total_additive_const_to_sd_mugal`, the iteration procedure
            failed and an assertion error is raised. This parameter is only considered when using the `additive`
            iteration approach.
        multiplicative_factor_step_size_percent : float, optional (default=10.0)
            Initial iteration step size for the multiplicative factor that is used to scale all setup
            observations. This parameter is only considered when using the `multiplicative` iteration approach.
        max_multiplicative_factor_to_sd_percent : float, optional (default=200.0)
            Maximum scaling factor when using the `multiplicative` iteration approach for scaling the SD of setup
            observations. Minimum = 100%.
        min_multiplicative_factor_to_sd_percent : float, optional (default=50.0)
            Minimum scaling factor when using the `multiplicative` iteration approach for scaling the SD of setup
            observations. Minimum = 1.0%, maximum = 100.0%
        drift_pol_degree : int, optional (default=1)
            Degree of estimated drift polynomial.
        sig0_mugal : int, optional (default=1)
            A priori standard deviation of unit weight of observations [µGal] for the stochastic model of the
            least-squares adjustment.
        scaling_factor_datum_observations : float, optional (default=1.0)
            Factor for scaling the standard deviation (SD) of g of datum stations. The scaled SD is is used for
            weighting the direct pseudo observations of g at the datum stations that are introduced as datum
            constraints.
        add_const_to_sd_of_observations_mugal : float, optional (default=0.0)
            The defined additive constant is added to the standard deviation (SD) of setup
            observations in order to scale the SD and the resulting weights to realistic values. In µGal. The scaling
            factor `scaling_factor_for_sd_of_observations` is applied before adding this constant!
        scaling_factor_for_sd_of_observations : float, optional (default=1.0)
            Scaling factor for the standard deviation of the setup observations. `add_const_to_sd_of_observations_mugal`
            is applied after applying the scaling factor!
        confidence_level_chi_test : float, optional (default=0.95)
            Confidence level for the goodness-of-fit test.
        confidence_level_tau_test : float, optional (default=0.95)
            Confidence level for the tau test.
        drift_ref_epoch_type : string ('survey' or 'campaign'), optional (default='survey')
            Defines whether the reference epoch t0 for the estimation of the drift polynomials for each survey in the
            campaign is the reference epoch of the first (active) observation in each survey (option: 'survey') or the
            first (active) observation in the whole campaign (option: 'campaign').
        noise_floor_mugal : float, optional (default=0.0)
            The standard error SE of the estimated gravity values at stations is calculated by SE = sqrt(SD**2 + NF**2),
            where SD is the estimated standard deviation of the station's gravity and NF is the noise floor value
            specified here.
        verbose : bool, optional (default=False)
            If `True`, status messages are printed to the command line, e.g. for debugging and testing
        """

        # Init.:
        flag_s0_within_threshold = False
        # scale_factor = scaling_factor_for_sd_of_observations
        add_const = add_const_to_sd_of_observations_mugal  # initial
        add_const_step_size = add_const_to_sd_of_observations_step_size_mugal  # [mugal]
        mult_factor = scaling_factor_for_sd_of_observations  # initial
        mult_factor_step_size = multiplicative_factor_step_size_percent / 100  # [%] => factor
        last_iteration_step_action = ''  # 'increase' or 'decrease'
        iteration_log_str = ''
        complete_log_str = ''
        add_const_total_mugal = 0
        mult_factor_total = 1.0

        # Run iteration:
        for i_iteration in range(1, max_number_iterations + 1):

            # Only apply multiplicative facot or additive constant in the first iteration, depending on the iteration
            # approach:
            if i_iteration > 1:
                if iteration_approach == 'Additive':
                    mult_factor = 1.0
                elif iteration_approach == 'Multiplicative':
                    add_const = 0.0

            add_const_total_mugal += add_const  # Log total additive constant
            mult_factor_total *= mult_factor  # Log total multiplicative factor
            self.log_str = ''  # Reset log string of the previous run

            self.adjust(drift_pol_degree=drift_pol_degree,
                        sig0_mugal=sig0_mugal,
                        scaling_factor_datum_observations=scaling_factor_datum_observations,
                        add_const_to_sd_of_observations_mugal=add_const,  # Adjusted iteratively
                        scaling_factor_for_sd_of_observations=mult_factor,  # Adjusted iteratively
                        confidence_level_chi_test=confidence_level_chi_test,
                        confidence_level_tau_test=confidence_level_tau_test,
                        drift_ref_epoch_type=drift_ref_epoch_type,
                        noise_floor_mugal=noise_floor_mugal,
                        verbose=False
                        )

            iteration_log_str_tmp = f'########## Iteration {i_iteration} #########################\n'
            # if iteration_approach == 'Additive':
            iteration_log_str_tmp += f'Total additive const. to SD: {add_const_total_mugal:5.3f} µGal\n'
            iteration_log_str_tmp += f'Current additive const. to SD: {add_const:5.3f} µGal\n'
            # elif iteration_approach == 'Multiplicative':
            iteration_log_str_tmp += f'Total mult. factor for SD: {mult_factor_total:5.3f}\n'
            iteration_log_str_tmp += f'Current mult. factor for SD: {mult_factor:5.3f}\n'
            iteration_log_str_tmp += f's0² a posteriori: {self.s02_a_posteriori:5.3f} µGal²\n'
            iteration_log_str_tmp += f'\n'
            if verbose:
                print(iteration_log_str_tmp)
                print(self.log_str)
            if self.write_log:
                complete_log_str = complete_log_str + iteration_log_str_tmp + self.log_str + '\n\n'
            iteration_log_str += iteration_log_str_tmp

            if (self.s02_a_posteriori < (s02_target + s02_target_delta)) and (
                    self.s02_a_posteriori > (s02_target - s02_target_delta)):
                flag_s0_within_threshold = True
            elif self.s02_a_posteriori > (s02_target + s02_target_delta):  # Too large => Increase SD of obs.
                if last_iteration_step_action == 'decrease':
                    add_const_step_size = add_const_step_size / 2
                    mult_factor_step_size = mult_factor_step_size / 2
                last_iteration_step_action = 'increase'
                add_const = add_const_step_size
                mult_factor = 1.0 + mult_factor_step_size
            elif self.s02_a_posteriori < (s02_target - s02_target_delta):  # Too small => Decrease SD of obs.
                if last_iteration_step_action == 'increase':
                    add_const_step_size = add_const_step_size / 2
                    mult_factor_step_size = mult_factor_step_size / 2
                last_iteration_step_action = 'decrease'
                add_const = -add_const_step_size
                mult_factor = 1.0 - mult_factor_step_size

            if flag_s0_within_threshold:  # Exit loop and stop iteration
                break

        # Write status message to log according to iteration result:
        if iteration_approach == 'Additive':
            if flag_s0_within_threshold and (add_const_total_mugal <= max_total_additive_const_to_sd_mugal):
                iteration_log_str_tmp = f' => Iteration successful!\n'
                iteration_log_str_tmp += f' => s0² a posteriori of {self.s02_a_posteriori:1.3f} within [{s02_target - s02_target_delta:1.3f}, {s02_target + s02_target_delta:1.3f}]\n'
                iteration_log_str_tmp += f' => Total additive constant to SD of observations ({add_const_total_mugal:1.3f}) ' + \
                                         f'is smaller than the user defined threshold ' + \
                                         f'of {max_total_additive_const_to_sd_mugal:1.3f} µGal.\n'
            else:
                iteration_log_str_tmp = f' => ERROR: Iteration failed!\n'
                if not flag_s0_within_threshold:
                    iteration_log_str_tmp += f' => s0² a posteriori of {self.s02_a_posteriori:1.3f} not within ' + \
                                             f'[{s02_target - s02_target_delta:1.3f}, {s02_target + s02_target_delta:1.3f}]\n'
                if (add_const_total_mugal > max_total_additive_const_to_sd_mugal):
                    iteration_log_str_tmp += f' => Total additive constant to SD  ({add_const_total_mugal:1.3f}) ' + \
                                             f'exceeds the the user defined threshold ' + \
                                             f'of {max_total_additive_const_to_sd_mugal:1.3f} µGal.\n'
            iteration_log_str_tmp += f'\n'

        elif iteration_approach == 'Multiplicative':
            if flag_s0_within_threshold and (mult_factor_total <= (max_multiplicative_factor_to_sd_percent / 100)) and (
                    mult_factor_total >= (min_multiplicative_factor_to_sd_percent / 100)):
                iteration_log_str_tmp = f' => Iteration successful!\n'
                iteration_log_str_tmp += f' => s0² a posteriori of {self.s02_a_posteriori:1.3f} within [{s02_target - s02_target_delta:1.3f}, {s02_target + s02_target_delta:1.3f}]\n '
                iteration_log_str_tmp += f' => Total multiplicative factor for SD of observations ({mult_factor_total * 100:1.3f}%) ' + \
                                         f'is between the user defined threshold ' + \
                                         f'of {min_multiplicative_factor_to_sd_percent:1.3f}% and {max_multiplicative_factor_to_sd_percent:1.3f}%.\n'
            else:
                iteration_log_str_tmp = f' => ERROR: Iteration failed!\n'
                if not flag_s0_within_threshold:
                    iteration_log_str_tmp += f' => s0² a posteriori of {self.s02_a_posteriori:1.3f} not within ' + \
                                             f'[{s02_target - s02_target_delta:1.3f}, {s02_target + s02_target_delta:1.3f}]\n'
                if (mult_factor_total > (max_multiplicative_factor_to_sd_percent / 100)):
                    iteration_log_str_tmp += f' => Total multiplicative factor for SD  ({mult_factor_total * 100:1.3f}%) ' + \
                                             f'exceeds the the user defined threshold ' + \
                                             f'of {max_multiplicative_factor_to_sd_percent:1.3f}%.\n'
                if (mult_factor_total < (min_multiplicative_factor_to_sd_percent / 100)):
                    iteration_log_str_tmp += f' => Total multiplicative factor for SD  ({mult_factor_total * 100:1.3f}%) ' + \
                                             f'is smaller than the user defined threshold ' + \
                                             f'of {min_multiplicative_factor_to_sd_percent:1.3f}%.\n'
            iteration_log_str_tmp += f'\n'

        iteration_log_str += iteration_log_str_tmp

        if verbose:
            print(iteration_log_str_tmp)

        # Append iteration log to the log string of the last iteration:
        if self.write_log:
            self.log_str = complete_log_str + \
                           '\n########## Iteration log #########################\n\n' + \
                           f' - Iteration approach: {iteration_approach}\n\n' + \
                           iteration_log_str

        if not flag_s0_within_threshold:
            raise AssertionError(f'Iteration Error: s0² a posteriori of {self.s02_a_posteriori:1.3f} not within '
                                 f'[{s02_target - s02_target_delta:1.3f}, {s02_target + s02_target_delta:1.3f}] after '
                                 f'{i_iteration} iterations!')

        if iteration_approach == 'Additive':
            if abs(add_const_total_mugal) > max_total_additive_const_to_sd_mugal:
                raise AssertionError(f'Total additive constant to SD  ({add_const_total_mugal:1.3f}) '
                                     f'exceeds the the user defined threshold '
                                     f'of {max_total_additive_const_to_sd_mugal:1.3f} µGal.\n')
        elif iteration_approach == 'Multiplicative':
            if mult_factor_total > (max_multiplicative_factor_to_sd_percent / 100):
                raise AssertionError(f'Total multiplicative factor for SD  ({mult_factor_total * 100:1.3f}%) '
                                     f'exceeds the the user defined threshold '
                                     f'of {max_multiplicative_factor_to_sd_percent:1.3f}%.\n')
            if mult_factor_total < (min_multiplicative_factor_to_sd_percent / 100):
                raise AssertionError(f'Total multiplicative factor for SD  ({mult_factor_total * 100:1.3f}%) '
                                     f'is smaller than the the user defined threshold '
                                     f'of {min_multiplicative_factor_to_sd_percent:1.3f}%.\n')

        self.number_of_iterations = i_iteration

    def export_stat_results_shapefile(self, filename, epsg_code, verbose=True):
        """Export station-related results from the `stat_obs_df` dataframe to a shapefile.

        Notes
        -----
        This method relies on the optional module `geopandas` (optional dependency).

        Parameters
        ---------
        filename : str
            Name and path of the output shapefile.
        epsg_code: int
            EPSG code of the station coordinates' CRS.
        verbose : bool, optional (default=False)
            `True` implies that status messages are printed to the command line.
        """
        if not _has_geopandas:
            raise ImportError(f'Optional dependency geopandas not available, but needed for writing shapefiles!')
        if verbose:
            print(f'Save station results of lsm run "{self.comment}" to: {filename}')

        stat_obs_df = self.stat_obs_df.copy(deep=True)
        stat_obs_df['comment'] = self.comment
        stat_obs_gdf = geopandas.GeoDataFrame(stat_obs_df, geometry=geopandas.points_from_xy(stat_obs_df['long_deg'],
                                                                                             stat_obs_df['lat_deg']),
                                              crs=epsg_code)
        stat_obs_gdf.to_file(filename)

    @property
    def time_str(self):
        """Return time of lsm adjustment as formatted string."""
        return self.init_time.strftime("%Y-%m-%d, %H:%M:%S")

    @property
    def get_log_string(self):
        """Returns the log string."""
        return self.log_str

    @property
    # @time_it
    def get_correlation_matrix(self):
        """Calculates und returns the correlation matrix based on the Co-Variance matrix of the LSM run.

        Returns
        -------
        numpy array : Correlation matrix (n, n)

        Notes
        -----
        The co-variance matrix `self.Cxx` needs to exist.
        """
        if self.Cxx is None:
            # raise AssertionError('Correlation matrix cannot be calculated due to missing co-variance matrix.')
            return None
        mat_Rxx = np.full((self.Cxx.shape[0], self.Cxx.shape[1]), np.nan)
        for i_row in range(0, mat_Rxx.shape[0]):
            cxx_row_sqrt = np.sqrt(self.Cxx[i_row, i_row])
            for i_col in range(0, i_row + 1):
                mat_Rxx[i_row, i_col] = self.Cxx[i_row, i_col] / (
                        cxx_row_sqrt * np.sqrt(self.Cxx[i_col, i_col]))
                mat_Rxx[i_col, i_row] = mat_Rxx[i_row, i_col]
        return mat_Rxx

    @property
    def get_results_obs_df(self):
        """Getter for the observation-related results."""
        return self.setup_obs_df

    @property
    def get_results_drift_df(self):
        """Getter for the drift-related results."""
        return self.drift_pol_df

    @property
    def get_results_stat_df(self):
        """Getter for the station-related results."""
        return self.stat_obs_df

    @property
    def get_results_vg_df(self):
        """Getter for the results of vertical gravity gradient estimation."""
        try:
            return self.vg_pol_df
        except AttributeError:  # If "self" has no attribute "vg_pol_df" (not initialized)
            return None

    @property
    def check_for_unique_setup_id(self):
        """Indicates whether setup IDs should be unique based on the setup data calculation method.

        Returns
        -------
        bool : `True`, if the setup IDs should be unique, otherwise `False`.

        Notes
        -----
        If observations were not aggregated within at least one setup, checking for unique setup IDs makes no sense.
        """
        for survey_name, setup_data in self.setups.items():
            if setup_data['setup_calc_method'] == 'individual_obs':
                return False
        return True

    def check_unique_setups(self, setup_ids: list):
        """Raise an error, is the setup IDs in the input list are not unique.

        Notes
        -----
        Omit the check, if `check_for_unique_setup_id` returns `False`.
        """
        if self.check_for_unique_setup_id:
            if len(set(setup_ids)) != len(setup_ids):
                non_unique_items = get_nonunique_items(setup_ids)
                non_unique_items_str = [str(item) for item in non_unique_items]
                non_unique_items_str = ", ".join(non_unique_items_str)
                raise RuntimeError(
                    f'Setup IDs are not unique within the campaign! Setups with the following IDs are not unique: {non_unique_items_str}')

    @property
    def survey_info_string(self) -> str:
        """Returns information on all surveys."""
        survey_names_list = []
        num_setups_list = []
        num_obs_total_list = []
        num_obs_active_list = []
        tide_correction_type_list = []
        reference_height_type_list = []
        atm_pres_correction_type_list = []
        scale_correction_type_list = []
        setup_calc_method_list = []
        for survey_name, survey in self.setups.items():
            survey_names_list.append(survey_name)
            setup_df = survey['setup_df']
            num_setups_list.append(len(setup_df))
            setup_obs_list_df = survey['setup_obs_list_df']
            num_obs_total_list.append(len(setup_obs_list_df))
            num_obs_active_list.append(len(setup_obs_list_df.loc[setup_obs_list_df['keep_obs']]))
            tide_correction_type_list.append(survey['tide_correction_type'])
            reference_height_type_list.append(survey['reference_height_type'])
            atm_pres_correction_type_list.append(survey['atm_pres_correction_type'])
            scale_correction_type_list.append(survey['scale_correction_type'])
            setup_calc_method_list.append(survey['setup_calc_method'])
        tmp_df = pd.DataFrame(list(zip(survey_names_list,
                                       num_setups_list,
                                       num_obs_total_list,
                                       num_obs_active_list,
                                       tide_correction_type_list,
                                       reference_height_type_list,
                                       atm_pres_correction_type_list,
                                       scale_correction_type_list,
                                       setup_calc_method_list,
                                       )),
                              columns=['Survey', '#Setups', '#Obs', '#Active obs', 'Tide correction', 'Ref. height',
                                       'Atm. correction', 'Scale correction', 'Setup calc. method'])
        return tmp_df.to_string(index=False)

    # @property
    # def obs_level_corrections_info_string(self) -> str:
    #     """Returns information on the applied observation-level corrections."""
    #     info_str = ''
    #     return info_str


def bin_redundancy_components(mat_r):
    """Bin redundancy components for easier interpretatzion.

    See: Skriptum AG2 (Navratil, TU Wien), p. 70.
    """
    number_obs_r_equal_0 = 0  # No error control: Errors cannot be detected
    number_obs_r_0_to_03 = 0  # Bad error control
    number_obs_r_03_to_07 = 0  # Good error control
    number_obs_r_07_1 = 0  # Very good error control, but observation may be redundant
    number_obs_r_equal_1 = 0  # Observation is redundant

    results_dict = {}  # '': ''

    return number_obs_r_equal_0, number_obs_r_0_to_03


def tau_test(mat_w, dof, alpha, mat_r):
    """Tau-criterion test for outlier detection of a least-squares adjustment.

    This test considers that the true variance factor sigma_0² is unknown. It is based on normalized residuals
    (Studentized residuals) and the Tau distribution.

    If the test fails it indicates that ONE (!) observation is a gross error, i.e. ONE residual is an outlier!
    If the adjustment is affected by two or more gross errors this test is not applicable. In that case the
    following pragmatic approach is recommended (e.g. by Caspary, 1987): The observation with the largest test
    statistic is discarded. Then the adjustment is repeated with the remaining n-1 observations and the test is
    repeated based on the new results, etc.

    Notes
    -----
    Ref.: Pope (1976): The statistics of residuals and the detection of outliers, NOAA Technical Report NOS 65 NGS 1
    Ref.: Caspari (1987): Concepts of network and deformation analysis. Monograph by UNSW Sydney, pp. 76-77
    This method implements a two-sided test, as the critical value is calculated with (alpha/2) and the abs. value of
    the normalized residual w is tested.

    Parameters
    ----------
    mat_w: np.array of floats
        Vector with standardized post-fit residuals of all observations in the least-squares adjustment.
        They are computed as post-fit residuals divides by their standard deviations.
    dof: int
        Degree of freedom of the least-squares adjustment.
    alpha: float (0 to 1)
        Significance level.
    mat_r: np.array of floats
        Vector with redundancy components derived in the least-squares adjustment for each observation.
    """
    # Critical value (two-sided test: the abs. value of w is tested und the Tau-distribution is symmetric about 0!):
    tsd_crt = stats.t.ppf(1 - alpha / 2, dof - 1)  # t distribution
    tau_crt = (tsd_crt * np.sqrt(dof)) / np.sqrt(dof - 1 + tsd_crt ** 2)  # Critical value (Pope, 1976, Eq. (6))
    tau_test_result = []
    for idx, w in enumerate(mat_w):
        # Check, if redancy component is larger than threshold:
        if mat_r[idx] > settings.R_POPE_TEST_THRESHOLD:
            if np.abs(w) > tau_crt:  # Equ. (6-36) in Caspary (1987)
                tau_test_result.append('failed')
            else:
                tau_test_result.append('passed')
        else:
            tau_test_result.append(f'r too small')  # Pope test not applied due to small redundancy component!
    return tau_test_result, tau_crt


def create_hist(mat_v):
    """Create histogram."""
    residuals = np.ndarray.tolist(mat_v)
    residuals = [item for sublist in residuals for item in sublist]
    hist_residuals, bin_edges = np.histogram(residuals, bins=5)
    return hist_residuals, bin_edges

def global_model_test(cf, dof, a_posteriori_variance_of_unit_weight, a_priori_variance_of_unit_weight):
    """Global model test based on the comparison of a posteriori and a priori variance of unit weight.

    This global model test asserts if the "model is correct and complete". If it fails it indicates that the
    observations contradict the mathematical adjustment model. However, this test cannot the validity of the model or
    the correctness of the observation!

    Notes
    -----
    This "global model test" is described by Caspari (1987): Concepts of network and deformation analysis. Monograph
    by UNSW Sydney, pp. 6-8 and pp.68-69.
    This method implements a two-sided test.

    Parameters
    ----------
    cf: float
        Confidence level for Chi²
    dof: int
        Degree of freedom.
    a_posteriori_variance_of_unit_weight: float
        A posteriori variance of unit weight (after adjustment).
    a_priori_variance_of_unit_weight: float
        A priori variance of unit weight (before adjustment).

    Returns
    -------
    """
    alpha = 1 - cf  # Significance level = Probability of committing a type 1 error (H0 wrongly dismissed)
    chi_crit_upper = stats.chi2.ppf(1 - alpha / 2, dof)  # critical value
    chi_crit_lower = stats.chi2.ppf(alpha / 2, dof)  # critical value
    # chi_critical_value = stats.chi2.ppf(alpha, dof)  # critical value
    test_value = dof * a_posteriori_variance_of_unit_weight / a_priori_variance_of_unit_weight  # Equ. (2-13) in Caspary (1987)
    if chi_crit_lower < test_value < chi_crit_upper:
        # if test_value <= test_value:  # Equ. (2-15) in Caspary (1987)
        chi_test_status = 'Passed'
    else:
        chi_test_status = 'Not passed'
    chi_crit = [chi_crit_lower, chi_crit_upper]
    return chi_crit, test_value, chi_test_status



"""Customized PyQt GUI elements for GravTools.

Copyright (C) 2023  Andreas Hellerschmied <andreas.hellerschmied@bev.gv.at>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""

from PyQt5.QtWidgets import QMessageBox, QScrollArea, QLabel, QGridLayout


class ScrollMessageBox(QMessageBox):
   """Custom PyQt message box with scroll area for large texts.

   Source:
   https://stackoverflow.com/questions/47345776/pyqt5-how-to-add-a-scrollbar-to-a-qmessagebox
   """
   def __init__(self, *args, **kwargs):
      QMessageBox.__init__(self, *args, **kwargs)
      chldn = self.children()
      scrll = QScrollArea(self)
      scrll.setWidgetResizable(True)
      grd = self.findChild(QGridLayout)
      lbl = QLabel(chldn[1].text(), self)
      lbl.setWordWrap(True)
      scrll.setWidget(lbl)
      scrll.setMinimumSize (400,200)
      grd.addWidget(scrll,0,1)
      chldn[1].setText('')
      self.exec_()


"""Models for surveys with the Scintrex CG-6 gravity meters.

This module contains all models that are specifically required to
handle observation with the Scintrex CG-5 gravity meter.

Copyright (C) 2024  Andreas Hellerschmied <andreas.hellerschmied@bev.gv.at>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import pandas as pd
import numpy as np
import re
import datetime as dt

from numpy import datetime64

from gravtools.models.exceptions import InvaliFileContentError
from gravtools import settings


class DataCursor:
    """Data cursor for matplotlib plot. X and Y coordinates are printed life to the plot canvas.

    From: https://stackoverflow.com/questions/4652439/is-there-a-matplotlib-equivalent-of-matlabs-datacursormode
    """
    # text_template = 'x: %0.2f\ny: %0.2f'
    x, y = 0.0, 0.0
    xoffset, yoffset = -20, 20
    text_template = 'x: %0.2f\ny: %0.2f'

    def __init__(self, ax):
        self.ax = ax
        self.annotation = ax.annotate(self.text_template,
                                      xy=(self.x, self.y), xytext=(self.xoffset, self.yoffset),
                                      textcoords='offset points', ha='right', va='bottom',
                                      bbox=dict(boxstyle='round,pad=0.5', fc='yellow', alpha=0.5),
                                      arrowprops=dict(arrowstyle='->', connectionstyle='arc3,rad=0')
                                      )
        self.annotation.set_visible(False)

    def __call__(self, event):
        self.event = event
        # xdata, ydata = event.artist.get_data()
        # self.x, self.y = xdata[event.ind], ydata[event.ind]
        self.x, self.y = event.mouseevent.xdata, event.mouseevent.ydata
        if self.x is not None:
            self.annotation.xy = self.x, self.y
            self.annotation.set_text(self.text_template % (self.x, self.y))
            self.annotation.set_visible(True)
            event.canvas.draw()


class CG6Survey:
    """CG-6 survey data.

    Beside the actual observations class instances contain alls the header information provided in the CG-6 observation
    files.

    Attributes
    ----------
    obs_filename : str
        Name (and path) of the CG-6 observation file.
    obs_file_type : str (`cg6_obs_file_lynx` or `cg6_obs_file_solo`)
        Observation file type as described and listed in `settings.CG6_SURVEY_DATA_SOURCE_TYPES`.
    survey_name : str
        Name of the survey.
    serial_number : str
        Gravimeter serial number.
    created_datetime : datetime object
        Creation time and date of the observation file.
    operator : str
        Instrument operator.
    gcal1 : float
        Instrumental calibration factor GCAL1 of the Scintrex CG6 gravimeter as used for the survey [mGal].
    goff : float
        Gravity offset of the Scintrex CG6 used for the survey [ADU].
    gref : float
        Gravity reference value of the Scintrex CG6 used for the survey [mGal].
    x_scale : float
        X scaling factor of the Scintrex CG6 used for the survey [arcsec/ADU].
    y_scale : float
        Y scaling factor of the Scintrex CG6 used for the survey [arcsec/ADU].
    x_offset : float
        X offset value of the Scintrex CG6 used for the survey [ADU].
    y_offset : float
        X offset value of the Scintrex CG6 used for the survey [ADU].
    temp_corr_coeff : float
        Temperature correction coefficient of the Scintrex CG6 used for the survey [mGal/mK].
    temp_sensor_scale : float
        Temperature sensor scale value of the Scintrex CG6 used for the survey [mK/ADU].
    drift_rate : float
        Long-term drift value [mGal/day]
    drift_ref_datetime : datetime object
        Drift zero time and date.
    tilt_corr : bool
        Flag that indicates whether tilt correction was enabled.
    tidal_corr : bool
        Flag that indicates whether tidal correction was enabled.
    ocean_loading_corr : bool
        Flag that indicates whether ocean loading correction was enabled. Only available when using LynxLG.
    temp_corr : bool
        Flag that indicates whether temperature correction was enabled.
    drift_corr : bool
        Flag that indicates whether drift correction was enabled.
    obs_df : pandas data frame
       Contains the actual observation data records with the columns defined in `self._OBS_DF_COLUMN_NAMES`.
    temp_sensor_offset : float, optional (default = numpy.nan)
        Temperature sensor offset. LynxLG only.
    firmware_version : str, optional (default = '')
        Firmware version of the instrument. Only available in CG6-solo observation files.
    gravity_filter_type : str, optional (default = '')
        Type of the filter used to calculate filtered gravity values. LynxLG only.
    gravity_filter_period : float, optional (default = numpy.nan)
        Length of filter period [s]. LynxLG only.
    date_time_source : str, optional (default = '')
        Source of date and time. LynxLG only.
    tidal_corr_model : str, optional (default = '')
        Name of the tidal correction model. LynxLG only.
    lynxlg_version : str, optional (default = '')
        LynxLG software version. LynxLG only.
    """
    # Column names of the dataframe containing tha actual observation data:
    _OBS_DF_COLUMNS = {
        'ref_time': np.datetime64,
        'station': str,
        'occupation': str,
        'line': str,
        'user_lat_deg': float,
        'user_lon_deg': float,
        'user_height_m': float,
        'g_corr_mugal': float,
        'se_mugal': float,
        'sd_mugal': float,
        'tilt_x_arcsec': float,
        'tilt_y_arcsec': float,
        'sensor_temp_mk': float,
        'corr_tilt_mugal': float,
        'corr_tide_mugal': float,
        'corr_oceanload_mugal': float,
        'corr_temp_mugal': float,
        'corr_drift_mugal': float,
        'gps_lat_deg': float,
        'gps_lon_deg': float,
        'gps_fix_quality': float,
        'gps_satellites': float,
        'gps_hdop': float,
        'gps_h_m': float,
        'duration': float,
        'num_rejected': float,
        'instr_height_m': float,
        'gradient_mugalm': float,
        'g_raw_mugal': float,
        'setup_id': float,
        'note': str,
    }
    _OBS_DF_COLUMN_NAMES = _OBS_DF_COLUMNS.keys()

    def __init__(self,
                 obs_filename: str,
                 obs_file_type: str,
                 survey_name: str,
                 serial_number: str,
                 created_datetime: dt.datetime,
                 operator: str,
                 gcal1: float,
                 goff: float,
                 gref: float,
                 x_scale: float,
                 y_scale: float,
                 x_offset: float,
                 y_offset: float,
                 temp_corr_coeff: float,
                 temp_sensor_scale: float,
                 drift_rate: float,
                 drift_ref_datetime: dt.datetime,
                 tilt_corr: bool,
                 tidal_corr: bool,
                 ocean_loading_corr: bool,
                 temp_corr: bool,
                 drift_corr: bool,
                 obs_df: pd.DataFrame,
                 temp_sensor_offset: float = np.nan,
                 firmware_version: str = '',
                 gravity_filter_type: str = '',
                 gravity_filter_period: float = np.nan,
                 date_time_source: str = '',
                 tidal_corr_model: str = '',
                 lynxlg_version: str = '',
                 ):
        """
        Parameters
        ----------
        obs_filename : str, optional (default = '')
            Name (and path) of CG-6 observation file.
        obs_file_type : str, optional (default = '')
            Observation file type as described and listed in `settings.CG6_SURVEY_DATA_SOURCE_TYPES`.
        survey_name : str
            Name of the survey.
        serial_number : str
            Gravimeter serial number.
        created_datetime : datetime object
            Creation time and date of the observation file.
        operator : str
            Instrument operator.
        gcal1 : float
            Instrumental calibration factor GCAL1 of the Scintrex CG6 gravimeter as used for the survey [mGal].
        goff : float
            Gravity offset of the Scintrex CG6 used for the survey [ADU].
        gref : float
            Gravity reference value of the Scintrex CG6 used for the survey [mGal].
        x_scale : float
            X scaling factor of the Scintrex CG6 used for the survey [arcsec/ADU].
        y_scale : float
            Y scaling factor of the Scintrex CG6 used for the survey [arcsec/ADU].
        x_offset : float
            X offset value of the Scintrex CG6 used for the survey [ADU].
        y_offset : float
            X offset value of the Scintrex CG6 used for the survey [ADU].
        temp_corr_coeff : float
            Temperature correction coefficient of the Scintrex CG6 used for the survey [mGal/mK].
        temp_sensor_scale : float
            Temperature sensor scale value of the Scintrex CG6 used for the survey [mK/ADU].
        drift_rate : float
            Long-term drift value [mGal/day]
        drift_ref_datetime : datetime object
            Drift zero time and date.
        tilt_corr : bool
            Flag that indicates whether tilt correction was enabled.
        tidal_corr : bool
            Flag that indicates whether tidal correction was enabled.
        ocean_loading_corr : bool
            Flag that indicates whether ocean loading correction was enabled. Only available when using LynxLG.
        temp_corr : bool
            Flag that indicates whether temperature correction was enabled.
        drift_corr : bool
            Flag that indicates whether drift correction was enabled.
        obs_df : pandas data frame
            Contains the actual observation data records with the columns defined in `self._OBS_DF_COLUMN_NAMES`.
        temp_sensor_offset : float, optional (default = numpy.nan)
            Temperature sensor offset. LynxLG only.
        firmware_version : str, optional (default = '')
            Firmware version of the instrument. Only available in CG6-solo observation files.
        gravity_filter_type : str, optional (default = '')
            Type of the filter used to calculate filtered gravity values. LynxLG only.
        gravity_filter_period : float, optional (default = numpy.nan)
            Length of filter period [s]. LynxLG only.
         date_time_source : str, optional (default = '')
            Source of date and time. LynxLG only.
        tidal_corr_model : str, optional (default = '')
            Name of the tidal correction model. LynxLG only.
        lynxlg_version : str, optional (default = '')
            LynxLG software version. LynxLG only.
        """
        # obs_filename
        if not isinstance(obs_filename, str):
            raise RuntimeError('"obs_filename" needs to be a string.')
        self.obs_filename = obs_filename

        # obs_file_type
        if obs_file_type not in settings.CG6_SURVEY_DATA_SOURCE_TYPES:
            raise RuntimeError(f'Invalid observation file type.')
        self.obs_file_type = obs_file_type

        # survey_name
        if not isinstance(survey_name, str):
            raise RuntimeError('"survey_name" needs to be a string.')
        self.survey_name = survey_name

        # serial_number
        if not isinstance(serial_number, str):
            raise RuntimeError('"serial_number" needs to be a string.')
        self.serial_number = serial_number

        # created_datetime
        if not isinstance(created_datetime, dt.datetime):
            raise RuntimeError('"created_datetime" needs to be a datetime object.')
        self.created_datetime = created_datetime

        # operator
        if not isinstance(operator, str):
            raise RuntimeError('"operator" needs to be a string.')
        self.operator = operator

        # gcal1
        if not isinstance(gcal1, float):
            raise RuntimeError('"gcal1" needs to be a float.')
        self.gcal1 = gcal1

        # goff
        if not isinstance(goff, float):
            raise RuntimeError('"goff" needs to be a float.')
        self.goff = goff

        # gref
        if not isinstance(gref, float):
            raise RuntimeError('"gref" needs to be a float.')
        self.gref = gref

        # x_scale
        if not isinstance(x_scale, float):
            raise RuntimeError('"x_scale" needs to be a float.')
        self.x_scale = x_scale

        # y_scale
        if not isinstance(y_scale, float):
            raise RuntimeError('"y_scale" needs to be a float.')
        self.y_scale = y_scale

        # x_offset
        if not isinstance(x_offset, float):
            raise RuntimeError('"x_offset" needs to be a float.')
        self.x_offset = x_offset

        # y_offset
        if not isinstance(y_offset, float):
            raise RuntimeError('"y_offset" needs to be a float.')
        self.y_offset = y_offset

        # temp_corr_coeff
        if not isinstance(temp_corr_coeff, float):
            raise RuntimeError('"temp_corr_coeff" needs to be a float.')
        self.temp_corr_coeff = temp_corr_coeff

        # temp_sensor_scale
        if not isinstance(temp_sensor_scale, float):
            raise RuntimeError('"temp_sensor_scale" needs to be a float.')
        self.temp_sensor_scale = temp_sensor_scale

        # drift_rate
        if not isinstance(drift_rate, float):
            raise RuntimeError('"drift_rate" needs to be a float.')
        self.drift_rate = drift_rate

        # drift_ref_datetime
        if not isinstance(drift_ref_datetime, dt.datetime):
            raise RuntimeError('"drift_ref_datetime" needs to be a datetime object.')
        self.drift_ref_datetime = drift_ref_datetime

        # tilt_corr
        if not isinstance(tilt_corr, bool):
            raise RuntimeError('"tilt_corr" needs to be a bool type variable.')
        self.tilt_corr = tilt_corr

        # tidal_corr
        if not isinstance(tidal_corr, bool):
            raise RuntimeError('"tidal_corr" needs to be a bool type variable.')
        self.tidal_corr = tidal_corr

        # ocean_loading_corr
        if not isinstance(ocean_loading_corr, bool):
            raise RuntimeError('"ocean_loading_corr" needs to be a bool type variable.')
        self.ocean_loading_corr = ocean_loading_corr

        # temp_corr
        if not isinstance(temp_corr, bool):
            raise RuntimeError('"temp_corr" needs to be a bool type variable.')
        self.temp_corr = temp_corr

        # drift_corr
        if not isinstance(drift_corr, bool):
            raise RuntimeError('"drift_corr" needs to be a bool type variable.')
        self.drift_corr = drift_corr

        # check obs_df:
        if not isinstance(obs_df, pd.DataFrame):
            raise RuntimeError('"obs_df" needs to be a pandas DataFrame.')
        self._check_obs_df_columns(obs_df)
        obs_df = self._obs_df_reorder_columns(obs_df)

        self.obs_df = obs_df

        # Optional parameters:

        # temp_sensor_offset
        if not isinstance(temp_sensor_offset, float) and (temp_sensor_offset is not np.nan):
            raise RuntimeError('"temp_sensor_offset" needs to be a float.')
        self.temp_sensor_offset = temp_sensor_offset

        # firmware_version
        if not isinstance(firmware_version, str) and (firmware_version != ''):
            raise RuntimeError('"firmware_version" needs to be a string.')
        self.firmware_version = firmware_version

        # gravity_filter_type
        if not isinstance(gravity_filter_type, str) and (gravity_filter_type != ''):
            raise RuntimeError('"gravity_filter_type" needs to be a string.')
        self.gravity_filter_type = gravity_filter_type

        # gravity_filter_period
        if not isinstance(gravity_filter_period, float) and (gravity_filter_period is not np.nan):
            raise RuntimeError('"gravity_filter_period" needs to be a float.')
        self.gravity_filter_period = gravity_filter_period

        # date_time_source
        if not isinstance(date_time_source, str) and (date_time_source != ''):
            raise RuntimeError('"date_time_source" needs to be a string.')
        self.date_time_source = date_time_source

        # tidal_corr_model
        if not isinstance(tidal_corr_model, str) and (tidal_corr_model != ''):
            raise RuntimeError('"tidal_corr_model" needs to be a string.')
        self.tidal_corr_model = tidal_corr_model

        # lynxlg_version
        if not isinstance(lynxlg_version, str) and (lynxlg_version != ''):
            raise RuntimeError('"lynxlg_version" needs to be a string.')
        self.lynxlg_version = lynxlg_version

    @staticmethod
    def read_cg6_file(filename, comment_marker: str = '#') -> tuple[str, list[str]]:
        """Returns the content of the input file as string and as a list of line strings.

        Notes
        -----
        Each line of the input file is returned with exactly one EOL symbol in the return string.

        Parameters
        ----------
        filename : str
            Name of the input file.
        comment_marker : str, optional (default = '#')
            All lines that begin with `comment_marker` are supposed to be comments and are skipped.

        Returns
        -------
            tuple[str, list[str]] : Content of the input file as one string and as list of line strings.
        """
        # Read in file and ignore comment lines:
        file_handle = open(filename, 'r')
        lines = [file_handle.readline().strip()]
        if lines[0].startswith('\ufeff'):
            # Check byte order mark and change codec. Reopen file if required with the correct codec.
            file_handle.close()
            file_handle = open(filename, 'r', encoding='utf-8-sig')
            lines = []
        for line in file_handle:
            line_tmp = line.strip()
            if not line_tmp.startswith(comment_marker):
                lines.append(line_tmp)
        file_handle.close()
        str_obs_file = '\n'.join(lines)

        # Remove all or add one end-of-line symbols from end of string:
        # - Last character of string has to be a \n so that regex works correctly!
        number_of_trailing_eol_symbols = 0
        str_idx = -1
        while str_obs_file[str_idx] == '\n':
            number_of_trailing_eol_symbols += 1
            str_idx -= 1
        if number_of_trailing_eol_symbols == 0:
            str_obs_file += '\n'
        elif number_of_trailing_eol_symbols > 1:
            str_obs_file = str_obs_file[0:str_idx + 2]

        return str_obs_file, lines

    @classmethod
    def from_lynxlg_file(cls, filename: str, obs_file_type: str, expect_notes: bool,
                         dt_setup_sec: float = None, verbose: bool = True):
        """Constractor for LynxLG formatted observation files (version 1 and version 2).

        Notes
        -----
        The difference to the V1 observation file format is, that the V2 format supports notes taken per setup. Notes
        are saved in the file right before the block of observations at a station. Be aware that the note for the first
        setup is written to the file header.

        Subsequent setups are supposed to be separated by note entries in the observation file or by station names.
        Optionally, a time span may be defined that is used to distinguish between consecutive setups. In the latter
        case observations are divided into different setups, if the time gap between them is larger than the defined
        value.

        Parameters
        ----------
        filename : str
            Name and path of the observation file.
        obs_file_type : str
            Type of observation file format. All valid identifiers are listed as keys in
            `settings.CG6_SURVEY_DATA_SOURCE_TYPES`.
        expect_notes : bool
            `True` implies that notes are mandatory in the observation file. Notes are only available in the LYnxLG
            format version 2.
        dt_setup_sec : float, optional (default = `None`)
            Observations are split up into separate setups, if the time gap between them is larger than the provided
            value (in seconds).
        verbose : bool, optional (default = `True`)
            True indicates that status messages are written to the command line.
        """

        _HEADER_LINES = {
            # <variable_name str>: [<regex_expr str>, <field_name str>, <flag_optional bool>, <default var>, <dtype>]
            'created_date': [r'\/ Date: (?P<input_str>\S+)\s*\n', 'Date', False, '', 'str'],
            'created_time': [r'\/ Time: (?P<input_str>\S+)\s*\n', 'Time', False, '', 'str'],
            'meter': [r'\/ Meter: (?P<input_str>\S+)\s*\n', 'Meter', False, '', 'str'],
            'operator': [r'\/ Operator: (?P<input_str>\S+)\s*\n', 'Operator', False, '', 'str'],
            'gcal1': [r'\/ GCAL1: (?P<input_str>\S+)\s*\n', 'GCAL1', False, np.nan, 'float'],
            'goff': [r'\/ Gravity Offset: (?P<input_str>\S+)\s*\n', 'Gravity Offset', False, np.nan, 'float'],
            'gref': [r'\/ Gravity Reference: (?P<input_str>\S+)\s*\n', 'Gravity Reference', False, np.nan, 'float'],
            'temp_sensor_scale': [r'\/ Sensor Temperature Gain: (?P<input_str>\S+)\s*\n', 'Sensor Temperature Gain',
                                  False, np.nan, 'float'],
            'temp_sensor_offset': [r'\/ Sensor Temperature Offset: (?P<input_str>\S+)\s*\n',
                                   'Sensor Temperature Offset', False, np.nan, 'float'],
            'x_scale': [r'\/ X Level Gain: (?P<input_str>\S+)\s*\n', 'X Level Gain', False, np.nan, 'float'],
            'x_offset': [r'\/ X Level Offset: (?P<input_str>\S+)\s*\n', 'X Level Offset', False, np.nan, 'float'],
            'y_scale': [r'\/ Y Level Gain: (?P<input_str>\S+)\s*\n', 'Y Level Gain', False, np.nan, 'float'],
            'y_offset': [r'\/ Y Level Offset: (?P<input_str>\S+)\s*\n', 'Y Level Offset', False, np.nan, 'float'],
            'gravity_filter_type': [r'\/ Gravity Filter: (?P<input_str>.+)\n', 'Gravity Filter', False, '', 'str'],
            'gravity_filter_period': [r'\/ Gravity Filter Period: (?P<input_str>\S+)\s*\n', 'Gravity Filter Period',
                                      False, np.nan, 'float'],
            'date_time_source': [r'\/ Date/Time Source: (?P<input_str>\S+)\s*\n', 'Date/Time Source', False, '', 'str'],
            'tilt_corr': [r'\/ Level Correction: (?P<input_str>\S+)\s*\n', 'Level Correction', False, '', 'str'],
            'tidal_corr': [r'\/ Tidal Correction: (?P<input_str>\S+)\s*\n', 'Tidal Correction', False, '', 'str'],
            'ocean_loading_corr': [r'\/ Ocean Load Correction: (?P<input_str>\S+)\s*\n', 'Ocean Load Correction', False,
                                   '', 'str'],
            'temp_corr': [r'\/ Temperature Correction: (?P<input_str>\S+)\s*\n', 'Temperature Correction', False, '',
                          'str'],
            'drift_corr': [r'\/ Drift Correction: (?P<input_str>\S+)\s*\n', 'Drift Correction', False, '', 'str'],
            'tidal_corr_model': [r'\/ Tidal Model: (?P<input_str>\S+)\s*\n', 'Tidal Model', False, '', 'str'],
            'temp_corr_coeff': [r'\/ Temperature Correction Coefficient \(mGals\/mK\): (?P<input_str>\S+)\s*\n',
                                'Temperature Correction Coefficient (mGals/mK)', False, np.nan, 'float'],
            'drift_rate': [r'\/ Drift Rate \(mGals\/day\): (?P<input_str>\S+)\s*\n', 'Drift Rate (mGals/day)', False,
                           np.nan, 'float'],
            'drift_ref_date': [r'\/ Drift Zero Date: (?P<input_str>\S+)\s*\n', 'Drift Zero Date', False, '', 'str'],
            'drift_ref_time': [r'\/ Drift Zero Time: (?P<input_str>\S+)\s*\n', 'Drift Zero Time', False, '', 'str'],
        }

        if verbose:
            print(f'Load CG6 observation file (LYnxLG, V2): {filename}')

        # Read file:
        file_str, lines = cls.read_cg6_file(filename)

        # Read header items:
        # - This item is treated separately, because it can occure more than once in the obs. file.
        expr = r'\/ Software Version: (?P<lynxlg_version>\S+)\s*\n'
        # Read setup parameters blocks from obs file string (only one block allowed!):
        lynxlg_version_list = []
        for block in re.finditer(expr, file_str):
            block_dict = block.groupdict()
            lynxlg_version_list.append(block_dict['lynxlg_version'])
        if len(set(lynxlg_version_list)) > 1:
            raise InvaliFileContentError(f'The file {filename} contains non-unique software version numbers!')
        lynxlg_version = lynxlg_version_list[0]

        # Read header strings:
        for key, item in _HEADER_LINES.items():
            expr = item[0]
            field_name = item[1]
            flag_optional = item[2]
            data_type = item[4]
            block_count = 0
            for block in re.finditer(expr, file_str):
                block_dict = block.groupdict()
                block_count += 1
            if block_count == 1:
                if data_type == 'str':
                    item[3] = block_dict['input_str']
                elif data_type == 'float':
                    item[3] = float(block_dict['input_str'])
                elif data_type == 'int':
                    item[3] = int(block_dict['input_str'])
                else:
                    raise RuntimeError(f'Unknown data type "{data_type}": Conversion of field "{field_name}" failed '
                                       f'when loading file {filename}.')
            elif block_count == 0:
                if not flag_optional:
                    raise InvaliFileContentError(f'The file {filename} does not contain a "{field_name}" header line.')
            if block_count > 1:
                raise InvaliFileContentError(f'The file {filename} contains more than one "{field_name}" header line.')
            # _HEADER_LINES[key] = item  # Not required due to pass by reference

        # Convert strings to bools:
        bool_lookup_dict = {'Enabled': True, 'Disabled': False}
        convert_to_bool_items = ['tilt_corr', 'tidal_corr', 'ocean_loading_corr', 'temp_corr', 'drift_corr']
        for item in convert_to_bool_items:
            header_str = _HEADER_LINES[item][3]
            try:
                _HEADER_LINES[item][3] = bool_lookup_dict[header_str]
            except KeyError:
                raise RuntimeError(f'Invalid value in header field "{_HEADER_LINES[item][1]}": {header_str}.')

        # Convert strings to datetime:
        datetime_str = _HEADER_LINES['drift_ref_date'][3] + ' ' + _HEADER_LINES['drift_ref_time'][3]
        drift_ref_datetime = dt.datetime.strptime(datetime_str, '%Y-%m-%d %H:%M:%S')
        datetime_str = _HEADER_LINES['created_date'][3] + ' ' + _HEADER_LINES['created_time'][3]
        created_datetime = dt.datetime.strptime(datetime_str, '%Y-%m-%d %H:%M:%S')

        # Read observations:
        ref_time_list = []
        station_list = []
        occupation_list = []
        line_list = []
        user_lat_deg_list = []
        user_lon_deg_list = []
        user_height_m_list = []
        g_corr_mugal_list = []
        se_mugal_list = []
        sd_mugal_list = []
        tilt_x_arcsec_list = []
        tilt_y_arcsec_list = []
        sensor_temp_mk_list = []
        corr_tilt_mugal_list = []
        corr_tide_mugal_list = []
        corr_oceanload_mugal_list = []
        corr_temp_mugal_list = []
        corr_drift_mugal_list = []
        gps_lat_deg_list = []
        gps_lon_deg_list = []
        gps_fix_quality_list = []
        gps_satellites_list = []
        gps_hdop_list = []
        gps_h_m_list = []
        duration_list = []
        num_rejected_list = []
        instr_height_m_list = []
        gradient_mugalm_list = []
        g_raw_mugal_list = []
        setup_id_list = []
        note_list = []
        survey_name_list = []

        if expect_notes:
            flag_first_note = False
            flag_new_setup = False
        else:
            flag_first_note = True
            flag_new_setup = True
            note = ''  # Reset note
        num_setups = 0
        num_obs = 0
        for line in lines:
            # print(line)
            if line.startswith('/ Notes:'):
                flag_new_setup = True
                flag_first_note = True
                note = line.split('/ Notes:')[1].strip()
            elif not line.startswith('/') and line and flag_first_note:  # obs. line
                expr = r'(?P<year>[0-9]{4}) (?P<month>[ 0-9][0-9]) (?P<day>[ 0-9][0-9]) (?P<hour>[ 0-9][0-9]) (?P<minute>[ 0-9][0-9]) (?P<second>[ 0-9][0-9]) (?P<millisecond>[ 0-9][ 0-9][0-9]) (?P<survey>\S+) (?P<station>\S+) (?P<occupation>\S*) (?P<line>\S*) (?P<user_lat_deg>-?[0-9.]+) (?P<user_lon_deg>-?[0-9.]+) (?P<user_height_m>-?[0-9.]+) (?P<corr_g_mgal>[0-9.]+) (?P<se_g_mgal>[0-9.]+) (?P<sd_g_mgal>[0-9.]+) (?P<tilt_x_arcsec>-?[0-9.]+) (?P<tilt_y_arcsec>-?[0-9.]+) (?P<sensor_temp_mk>-?[0-9.]+) (?P<corr_tilt_mgal>-?[0-9.]+) (?P<corr_tide_mgal>-?[0-9.]+) (?P<corr_oceanload_mgal>-?[0-9.]+) (?P<corr_temp_mgal>-?[0-9.]+) (?P<corr_drift_mgal>-?[0-9.]+) (?P<gps_lat_deg>-?[0-9.]+) (?P<gps_lon_deg>-?[0-9.]+) +(?P<gps_fix_quality>[0-9]+) +(?P<gps_satellites>[0-9]+) (?P<gps_hdop>-?[0-9.]+) (?P<gps_h_m>-?[0-9.]+) +(?P<duration>[0-9]+) +(?P<num_rejected>[0-9]+) (?P<instr_height_m>[0-9.]+) (?P<gradient_mgalcm>-?[0-9.]+)'
                block_count = 0
                for block in re.finditer(expr, line):
                    block_dict = block.groupdict()
                    block_count += 1
                if block_count == 0:
                    raise InvaliFileContentError(
                        f'Invalid observation line in file {filename}: {line}. Could not match with regex.')
                elif block_count > 1:
                    raise InvaliFileContentError(
                        f'Invalid observation line in file {filename}: {line}. Multiple matches with regex.')

                # One match => OK:
                num_obs += 1

                t_ref = dt.datetime(year=int(block_dict['year']), month=int(block_dict['month']),
                                    day=int(block_dict['day']),
                                    hour=int(block_dict['hour']), minute=int(block_dict['minute']),
                                    second=int(block_dict['second']), microsecond=int(block_dict['millisecond']) * 1000)
                ref_time_list.append(t_ref)

                station_list.append(block_dict['station'])
                occupation_list.append(block_dict['occupation'])
                line_list.append(block_dict['line'])
                user_lat_deg_list.append(float(block_dict['user_lat_deg']))
                user_lon_deg_list.append(float(block_dict['user_lon_deg']))
                user_height_m_list.append(float(block_dict['user_height_m']))
                g_corr_mugal_list.append(float(block_dict['corr_g_mgal']) * 1e3)
                se_mugal_list.append(float(block_dict['se_g_mgal']) * 1e3)
                sd_mugal_list.append(float(block_dict['sd_g_mgal']) * 1e3)
                tilt_x_arcsec_list.append(float(block_dict['tilt_x_arcsec']))
                tilt_y_arcsec_list.append(float(block_dict['tilt_y_arcsec']))
                sensor_temp_mk_list.append(float(block_dict['sensor_temp_mk']))
                corr_tilt_mugal_list.append(float(block_dict['corr_tilt_mgal']) * 1e3)
                corr_tide_mugal_list.append(float(block_dict['corr_tide_mgal']) * 1e3)
                corr_oceanload_mugal_list.append(float(block_dict['corr_oceanload_mgal']) * 1e3)
                corr_temp_mugal_list.append(float(block_dict['corr_temp_mgal']) * 1e3)
                corr_drift_mugal_list.append(float(block_dict['corr_drift_mgal']) * 1e3)
                gps_lat_deg_list.append(float(block_dict['gps_lat_deg']))
                gps_lon_deg_list.append(float(block_dict['gps_lon_deg']))
                gps_fix_quality_list.append(float(block_dict['gps_fix_quality']))
                gps_satellites_list.append(float(block_dict['gps_satellites']))
                gps_hdop_list.append(float(block_dict['gps_hdop']))
                gps_h_m_list.append(float(block_dict['gps_h_m']))
                duration_list.append(float(block_dict['duration']))
                num_rejected_list.append(float(block_dict['num_rejected']))
                instr_height_m_list.append(float(block_dict['instr_height_m']))
                gradient_mugalm_list.append(float(block_dict[
                                                      'gradient_mgalcm']) * 1e5)  # TODO: Unit mgal/cm makes no sense...
                g_raw_mugal_list.append(np.nan)
                survey_name_list.append(block_dict['survey'])

                # Check conditions for new setups:
                if num_obs > 1:  # Minimum 2nd observation in survey
                    if station_list[num_obs - 1] != station_list[num_obs - 2]:  # Check station names to create setups
                        flag_new_setup = True
                        note = ''  # Reset note
                    elif dt_setup_sec is not None:  # Check time gap
                        if (ref_time_list[num_obs - 1] - ref_time_list[num_obs - 2]).seconds > dt_setup_sec:
                            flag_new_setup = True
                            note = ''  # Reset note

                # # Check time gap between observations to create setups:
                # if dt_setup_sec is not None:
                #     if not flag_new_setup:  # At least one obs. after a "/ Notes" line
                #         if num_obs > 0:
                #             if (ref_time_list[num_obs - 1] - ref_time_list[num_obs - 2]).seconds > dt_setup_sec:
                #                 flag_new_setup = True
                #                 note = ''  # Reset note

                # New setup id:
                if flag_new_setup:
                    flag_new_setup = False
                    setup_id = int(dt.datetime.timestamp(t_ref))
                    num_setups += 1
                setup_id_list.append(setup_id)
                note_list.append(note)

        # Check, whether data was loaded from file:
        if num_obs == 0:
            raise RuntimeError(f'No observations loaded from CG6 file ({filename}). Check, whether the correct file '
                               f'format was selected!')

        # Get survey name from observation lines:
        if len(set(survey_name_list)) > 1:
            raise RuntimeError(f'The survey name is not unique in the observation file {filename}!')
        survey_name = set(survey_name_list).pop()

        # Create pandas dataframe:
        obs_df = pd.DataFrame(list(zip(ref_time_list,
                                       station_list,
                                       occupation_list,
                                       line_list,
                                       user_lat_deg_list,
                                       user_lon_deg_list,
                                       user_height_m_list,
                                       g_corr_mugal_list,
                                       se_mugal_list,
                                       sd_mugal_list,
                                       tilt_x_arcsec_list,
                                       tilt_y_arcsec_list,
                                       sensor_temp_mk_list,
                                       corr_tilt_mugal_list,
                                       corr_tide_mugal_list,
                                       corr_oceanload_mugal_list,
                                       corr_temp_mugal_list,
                                       corr_drift_mugal_list,
                                       gps_lat_deg_list,
                                       gps_lon_deg_list,
                                       gps_fix_quality_list,
                                       gps_satellites_list,
                                       gps_hdop_list,
                                       gps_h_m_list,
                                       duration_list,
                                       num_rejected_list,
                                       instr_height_m_list,
                                       gradient_mugalm_list,
                                       g_raw_mugal_list,
                                       setup_id_list,
                                       note_list)),
                              columns=cls._OBS_DF_COLUMN_NAMES)

        return cls(obs_filename=filename,
                   obs_file_type=obs_file_type,
                   survey_name=survey_name,  # From observation block. Check for equality!
                   serial_number=_HEADER_LINES['meter'][3],
                   created_datetime=created_datetime,
                   operator=_HEADER_LINES['operator'][3],
                   gcal1=_HEADER_LINES['gcal1'][3],
                   goff=_HEADER_LINES['goff'][3],
                   gref=_HEADER_LINES['gref'][3],
                   x_scale=_HEADER_LINES['x_scale'][3],
                   y_scale=_HEADER_LINES['y_scale'][3],
                   x_offset=_HEADER_LINES['x_offset'][3],
                   y_offset=_HEADER_LINES['y_offset'][3],
                   temp_corr_coeff=_HEADER_LINES['temp_corr_coeff'][3],
                   temp_sensor_scale=_HEADER_LINES['temp_sensor_scale'][3],
                   drift_rate=_HEADER_LINES['drift_rate'][3],
                   drift_ref_datetime=drift_ref_datetime,
                   tilt_corr=_HEADER_LINES['tilt_corr'][3],
                   tidal_corr=_HEADER_LINES['tidal_corr'][3],
                   ocean_loading_corr=_HEADER_LINES['ocean_loading_corr'][3],
                   temp_corr=_HEADER_LINES['temp_corr'][3],
                   drift_corr=_HEADER_LINES['drift_corr'][3],
                   obs_df=obs_df,
                   temp_sensor_offset=_HEADER_LINES['temp_sensor_offset'][3],
                   # firmware_version: str, optional(default='')
                   gravity_filter_type=_HEADER_LINES['gravity_filter_type'][3],
                   gravity_filter_period=_HEADER_LINES['gravity_filter_period'][3],
                   date_time_source=_HEADER_LINES['date_time_source'][3],
                   tidal_corr_model=_HEADER_LINES['tidal_corr_model'][3],
                   lynxlg_version=lynxlg_version)

    @classmethod
    def from_lynxlg_file_v1(cls, filename: str, dt_setup_sec: float = None, verbose: bool = True):
        """Constractor for LynxLG formatted observation files (version 1, without notes).

        Notes
        -----
        The difference to the V1 observation file format is, that the V2 format supports notes taken per setup. Notes
        are saved in the file right before the block of observations at a station. Be aware that the note for the first
        setup is written to the file header.

        Subsequent setups at the same station can be separated based on the time difference between observations. If the
        separation of the reference time of two observations is larger than the value given by `dt_setup_sec` in seconds
        they are assumed to belong to two consecutive setups.

        Parameters
        ----------
        filename : str
            Name and path of the observation file.
        dt_setup_sec : float, optional (default = `None`)
            Minimum time gap between two consecutive observations [sec] in order to create separate setups. `None`
            implies that time gaps are not considered in order to create setups.
        verbose : bool, optional (default = True)
            True indicates that status messages are written to the command line.
        """
        return cls.from_lynxlg_file(filename, expect_notes=False, obs_file_type='cg6_obs_file_lynx_v1',
                                    dt_setup_sec=dt_setup_sec, verbose=verbose)

    @classmethod
    def from_lynxlg_file_v2(cls, filename: str, dt_setup_sec: float = None, verbose: bool = True):
        """Constractor for LynxLG formatted observation files (version 2, with notes).

        Notes
        -----
        The difference to the V1 observation file format is, that the V2 format supports notes taken per setup. Notes
        are saved in the file right before the block of observations at a station. Be aware that the note for the first
        setup is written to the file header.

        Subsequent setups are supposed to be separated by note entries in the observation file and by station names.
        Optionally, a time span may be defined that is used to distinguish between consecutive setups. In the latter
        case observations are divided into different setups, if the time gap between them is larger than the defined
        value.

        Parameters
        ----------
        filename : str
            Name and path of the observation file.
        dt_setup_sec : float, optional (default = `None`)
            Minimum time gap between two consecutive observations [sec] in order to create separate setups. `None`
            implies that time gaps are not considered in order to create setups.
        verbose : bool, optional (default = `True`)
            True indicates that status messages are written to the command line.
        """
        return cls.from_lynxlg_file(filename, obs_file_type='cg6_obs_file_lynx_v2', expect_notes=True,
                                    dt_setup_sec=dt_setup_sec, verbose=verbose)

    @classmethod
    def from_cg6solo_file(cls, filename: str, dt_setup_sec: float = None, verbose: bool = True):
        """Constractor for CG6-solo formatted observation files (version 2).

        Notes
        -----
        Optionally, subsequent setups at the same station can be separated based on the time difference between
        observations. If the separation of the reference time of two observations is larger than the value given by
        `dt_setup_sec` in seconds they are assumed to belong to two consecutive setups.

        Parameters
        ----------
        filename : str
            Name and path of the observation file.
        dt_setup_sec : float, optional (default = `None`)
            Minimum time gap between two consecutive observations [sec] in order to create separate setups. `None`
            implies that time gaps are not considered in order to create setups.
        verbose : bool, optional (default = True)
            True indicates that status messages are written to the command line.
        """
        _HEADER_LINES = {
            # <variable_name str>: [<regex_expr str>, <field_name str>, <flag_optional bool>, <default var>, <dtype>]
            'survey_name': [r'\/\t\tSurvey Name:\t(?P<input_str>\S+)\n', 'Survey name', False, '', 'str'],
            'serial_number': [r'\/\t\tInstrument Serial Number:\t(?P<input_str>\S+)\n', 'Instrument Serial Number',
                              False, '', 'str'],
            'created_datetime': [r'\/\t\tCreated:\t(?P<input_str>[0-9 :-]+)\n', 'Created', False, '', 'str'],
            'operator': [r'\/\t\tOperator:\t(?P<input_str>\S+)\n', 'Operator', False, '', 'str'],
            'gcal1': [r'\/\t\tGcal1 \[mGal\]:\t(?P<input_str>\S+)\n', 'Gcal1 [mGal]', False, np.nan, 'float'],
            'goff': [r'\/\t\tGoff \[ADU\]:\t(?P<input_str>\S+)\n', 'Goff [ADU]', False, np.nan, 'float'],
            'gref': [r'\/\t\tGref \[mGal\]:\t(?P<input_str>\S+)\n', 'Gref [mGal]', False, np.nan, 'float'],
            'x_scale': [r'\/\t\tX Scale \[arc-sec\/ADU\]:\t(?P<input_str>\S+)\n', 'X Scale [arc-sec/ADU]', False,
                        np.nan, 'float'],
            'y_scale': [r'\/\t\tY Scale \[arc-sec\/ADU\]:\t(?P<input_str>\S+)\n', 'Y Scale [arc-sec/ADU]', False,
                        np.nan, 'float'],
            'x_offset': [r'\/\t\tX Offset \[ADU\]:\t(?P<input_str>\S+)\n', 'X Offset [ADU]', False, np.nan, 'float'],
            'y_offset': [r'\/\t\tY Offset \[ADU\]:\t(?P<input_str>\S+)\n', 'Y Offset [ADU]', False, np.nan, 'float'],
            'temp_corr_coeff': [r'\/\t\tTemperature Coefficient \[mGal\/mK\]:\t(?P<input_str>\S+)\n',
                                'Temperature Coefficient [mGal/mK]', False, np.nan, 'float'],
            'temp_sensor_scale': [r'\/\t\tTemperature Scale \[mK\/ADU\]:\t(?P<input_str>\S+)\n',
                                  'Temperature Scale [mK/ADU]',
                                  False, np.nan, 'float'],
            'drift_rate': [r'\/\t\tDrift Rate \[mGal\/day\]:\t(?P<input_str>\S+)\n', 'Drift Rate [mGal/day]', False,
                           np.nan, 'float'],
            'drift_ref_datetime': [r'\/\t\tDrift Zero Time:\t(?P<input_str>[0-9 :-]+)\n', 'Drift Zero Time', False, '',
                                   'str'],
            'firmware_version': [r'\/\t\tFirmware Version:\t(?P<input_str>\S+)\n', 'Firmware Version', False, '', 'str'],
        }

        if verbose:
            print(f'Load CG6 observation file (CG6 solo): {filename}')
            if dt_setup_sec is not None:
                print(f' - A time gap of {dt_setup_sec} seconds or more between observations creates a new setup.')

        # Read file:
        file_str, lines = cls.read_cg6_file(filename)

        # Read header strings:
        for key, item in _HEADER_LINES.items():
            expr = item[0]
            field_name = item[1]
            flag_optional = item[2]
            data_type = item[4]
            block_count = 0
            for block in re.finditer(expr, file_str):
                block_dict = block.groupdict()
                block_count += 1
            if block_count == 1:
                if data_type == 'str':
                    item[3] = block_dict['input_str']
                elif data_type == 'float':
                    item[3] = float(block_dict['input_str'])
                elif data_type == 'int':
                    item[3] = int(block_dict['input_str'])
                else:
                    raise RuntimeError(
                        f'Unknown data type "{data_type}": Conversion of field "{field_name}" failed '
                        f'when loading file {filename}.')
            elif block_count == 0:
                if not flag_optional:
                    raise InvaliFileContentError(
                        f'The file {filename} does not contain a "{field_name}" header line.')
            if block_count > 1:
                raise InvaliFileContentError(
                    f'The file {filename} contains more than one "{field_name}" header line.')
            # _HEADER_LINES[key] = item  # Not required due to pass by reference

        # Convert strings to datetime:
        drift_ref_datetime = dt.datetime.strptime(_HEADER_LINES['drift_ref_datetime'][3], '%Y-%m-%d %H:%M:%S')
        created_datetime = dt.datetime.strptime(_HEADER_LINES['created_datetime'][3], '%Y-%m-%d %H:%M:%S')

        # Read observations:
        station_list = []
        ref_time_list = []
        g_corr_mugal_list = []
        line_list = []
        sd_mugal_list = []
        se_mugal_list = []
        g_raw_mugal_list = []
        tilt_x_arcsec_list = []
        tilt_y_arcsec_list = []
        sensor_temp_mk_list = []
        corr_tide_mugal_list = []
        corr_tilt_mugal_list = []
        corr_temp_mugal_list = []
        corr_drift_mugal_list = []
        duration_list = []
        instr_height_m_list = []
        user_lat_deg_list = []
        user_lon_deg_list = []
        user_height_m_list = []
        gps_lat_deg_list = []
        gps_lon_deg_list = []
        gps_h_m_list = []
        correction_flags_list = []
        occupation_list = []
        corr_oceanload_mugal_list = []
        gradient_mugalm_list = []
        gps_fix_quality_list = []
        gps_satellites_list = []
        gps_hdop_list = []
        num_rejected_list = []
        note_list = []
        setup_id_list = []

        # expr = r'(?P<station>\S+)\t(?P<date>[0-9-]{10})\t(?P<time>[0-9:]{8})\t(?P<corr_g_mgal>[0-9.]+)\t(?P<line>\S*)\t(?P<sd_g_mgal>[0-9.]+)\t(?P<se_g_mgal>[0-9.]+)\t(?P<raw_g_mgal>[0-9.]+)\t(?P<tilt_x_arcsec>-?[0-9.]+)\t(?P<tilt_y_arcsec>-?[0-9.]+)\t(?P<sensor_temp_mk>-?[0-9.]+)\t(?P<corr_tide_mgal>-?[0-9.]+)\t(?P<corr_tilt_mgal>-?[0-9.]+)\t(?P<corr_temp_mgal>-?[0-9.]+)\t(?P<corr_drift_mgal>-?[0-9.]+)\t(?P<duration>[0-9]+)\t(?P<instr_height_m>-?[0-9.]+)\t(?P<user_lat_deg>-?[0-9.]+)\t(?P<user_lon_deg>-?[0-9.]+)\t(?P<user_height_m>-?[0-9.]+)\t(?P<gps_lat_deg>-?[0-9.]+)\t(?P<gps_lon_deg>-?[0-9.]+)\t(?P<gps_height_m>-?[0-9.]+)\t(?P<correction_flags>[01]{5})\s*'
        expr = r'(?P<station>\S+)\t(?P<date>[0-9-]{10})\t(?P<time>[0-9:]{8})\t(?P<corr_g_mgal>[0-9.]+)\t(?P<line>\S*)\t(?P<sd_g_mgal>[0-9.]+)\t(?P<se_g_mgal>[0-9.]+)\t(?P<raw_g_mgal>[0-9.]+)\t(?P<tilt_x_arcsec>-?[0-9.]+)\t(?P<tilt_y_arcsec>-?[0-9.]+)\t(?P<sensor_temp_mk>-?[0-9.]+)\t(?P<corr_tide_mgal>-?[0-9.]+)\t(?P<corr_tilt_mgal>-?[0-9.]+)\t(?P<corr_temp_mgal>-?[0-9.]+)\t(?P<corr_drift_mgal>-?[0-9.]+)\t(?P<duration>[0-9]+)\t(?P<instr_height_m>-?[0-9.]+)\t(?P<user_lat_deg>-?[0-9.]+)\t(?P<user_lon_deg>-?[0-9.]+)\t(?P<user_height_m>-?[0-9.]+)\t(?P<gps_lat_deg>-?[0-9.-]+)\t(?P<gps_lon_deg>-?[0-9.-]+)\t(?P<gps_height_m>-?[0-9.-]+)\t(?P<correction_flags>[01]{5})\s*'
        obs_count = 0
        for block in re.finditer(expr, file_str):
            block_dict = block.groupdict()
            station_list.append(block_dict['station'])
            datetime_str = block_dict['date'] + ' ' + block_dict['time']
            ref_time_list.append(dt.datetime.strptime(datetime_str, '%Y-%m-%d %H:%M:%S'))
            g_corr_mugal_list.append(float(block_dict['corr_g_mgal']) * 1000)
            line_list.append(block_dict['line'])
            sd_mugal_list.append(float(block_dict['sd_g_mgal']) * 1000)
            se_mugal_list.append(float(block_dict['se_g_mgal']) * 1000)
            g_raw_mugal_list.append(float(block_dict['raw_g_mgal']) * 1000)
            tilt_x_arcsec_list.append(float(block_dict['tilt_x_arcsec']))
            tilt_y_arcsec_list.append(float(block_dict['tilt_y_arcsec']))
            sensor_temp_mk_list.append(float(block_dict['sensor_temp_mk']))
            corr_tide_mugal_list.append(float(block_dict['corr_tide_mgal']) * 1000)
            corr_tilt_mugal_list.append(float(block_dict['corr_tilt_mgal']) * 1000)
            corr_temp_mugal_list.append(float(block_dict['corr_temp_mgal']) * 1000)
            corr_drift_mugal_list.append(float(block_dict['corr_drift_mgal']) * 1000)
            duration_list.append(float(block_dict['duration']))  # TODO: [sec]?
            instr_height_m_list.append(float(block_dict['instr_height_m']))
            user_lat_deg_list.append(float(block_dict['user_lat_deg']))
            user_lon_deg_list.append(float(block_dict['user_lon_deg']))
            user_height_m_list.append(float(block_dict['user_height_m']))
            try:
                gps_lat_deg_list.append(float(block_dict['gps_lat_deg']))
                gps_lon_deg_list.append(float(block_dict['gps_lon_deg']))
                gps_h_m_list.append(float(block_dict['gps_height_m']))
            except ValueError:  # In case no GPS signal was available/recorded, e.g. at indoor use
                gps_lat_deg_list.append(float('nan'))
                gps_lon_deg_list.append(float('nan'))
                gps_h_m_list.append(float('nan'))
            correction_flags_list.append(block_dict['correction_flags'])
            # Initialize fields that are not available in the obs. file:
            occupation_list.append('')
            corr_oceanload_mugal_list.append(np.nan)
            gradient_mugalm_list.append(np.nan)
            gps_fix_quality_list.append(np.nan)
            gps_satellites_list.append(np.nan)
            gps_hdop_list.append(np.nan)
            num_rejected_list.append(np.nan)
            note_list.append('')

            obs_count += 1
        if obs_count == 0:
            raise RuntimeError(f'No observations found in file {filename}.')

        # Separate instrument setups based on the time gaps and station names and determine setup IDs:
        num_setups = 0
        flag_new_setup = False
        for count_idx, ref_time in enumerate(ref_time_list):
            if count_idx == 0:  # First observation
                flag_new_setup = True
            else:
                if station_list[count_idx] != station_list[count_idx - 1]:  # New station?
                    flag_new_setup = True
                elif dt_setup_sec is not None:
                    if (ref_time_list[count_idx] - ref_time_list[count_idx - 1]).seconds > dt_setup_sec:  # Time gap?
                        flag_new_setup = True
            if flag_new_setup:
                setup_id = int(dt.datetime.timestamp(ref_time))
                num_setups += 1
                flag_new_setup = False
            setup_id_list.append(setup_id)

        # Create pandas dataframe:
        obs_df = pd.DataFrame(list(zip(ref_time_list,
                                       station_list,
                                       occupation_list,
                                       line_list,
                                       user_lat_deg_list,
                                       user_lon_deg_list,
                                       user_height_m_list,
                                       g_corr_mugal_list,
                                       se_mugal_list,
                                       sd_mugal_list,
                                       tilt_x_arcsec_list,
                                       tilt_y_arcsec_list,
                                       sensor_temp_mk_list,
                                       corr_tilt_mugal_list,
                                       corr_tide_mugal_list,
                                       corr_oceanload_mugal_list,
                                       corr_temp_mugal_list,
                                       corr_drift_mugal_list,
                                       gps_lat_deg_list,
                                       gps_lon_deg_list,
                                       gps_fix_quality_list,
                                       gps_satellites_list,
                                       gps_hdop_list,
                                       gps_h_m_list,
                                       duration_list,
                                       num_rejected_list,
                                       instr_height_m_list,
                                       gradient_mugalm_list,
                                       g_raw_mugal_list,
                                       setup_id_list,
                                       note_list)),
                              columns=cls._OBS_DF_COLUMN_NAMES)

        # Handle correction flags:
        if not all(item == correction_flags_list[0] for item in correction_flags_list):
            raise RuntimeError(f'The correction flags in the file {filename} are not equal for all observations!')
        drift_corr = correction_flags_list[0][0] == '1'
        temp_corr = correction_flags_list[0][1] == '1'
        tidal_corr = correction_flags_list[0][3] == '1'
        tilt_corr = correction_flags_list[0][4] == '1'

        return cls(obs_filename=filename,
                   obs_file_type='cg6_obs_file_solo',
                   survey_name=_HEADER_LINES['survey_name'][3],
                   serial_number=_HEADER_LINES['serial_number'][3],
                   created_datetime=created_datetime,
                   operator=_HEADER_LINES['operator'][3],
                   gcal1=_HEADER_LINES['gcal1'][3],
                   goff=_HEADER_LINES['goff'][3],
                   gref=_HEADER_LINES['gref'][3],
                   x_scale=_HEADER_LINES['x_scale'][3],
                   y_scale=_HEADER_LINES['y_scale'][3],
                   x_offset=_HEADER_LINES['x_offset'][3],
                   y_offset=_HEADER_LINES['y_offset'][3],
                   temp_corr_coeff=_HEADER_LINES['temp_corr_coeff'][3],
                   temp_sensor_scale=_HEADER_LINES['temp_sensor_scale'][3],
                   drift_rate=_HEADER_LINES['drift_rate'][3],
                   drift_ref_datetime=drift_ref_datetime,
                   tilt_corr=tilt_corr,
                   tidal_corr=tidal_corr,
                   ocean_loading_corr=False,
                   temp_corr=temp_corr,
                   drift_corr=drift_corr,
                   obs_df=obs_df,
                   # temp_sensor_offset=_HEADER_LINES['temp_sensor_offset'][3],  # Not available
                   firmware_version=_HEADER_LINES['firmware_version'][3]
                   # gravity_filter_type=_HEADER_LINES['gravity_filter_type'][3],  # Not available
                   # gravity_filter_period=_HEADER_LINES['gravity_filter_period'][3],  # Not available
                   # date_time_source=_HEADER_LINES['date_time_source'][3],  # Not available
                   # tidal_corr_model=_HEADER_LINES['tidal_corr_model'][3],  # Not available
                   # lynxlg_version=lynxlg_version  # Not available
                   )

    @classmethod
    def _check_obs_df_columns(cls, obs_df):
        """Check the columns of the `obs_df` DataFrame."""
        # Check for missing columns:
        missing_columns = set(cls._OBS_DF_COLUMN_NAMES) - set(obs_df.columns)
        if len(missing_columns):
            col_names = ', '.join(i for i in missing_columns)
            raise RuntimeError(f'Missing columns in CG6Survey.obs_df: {col_names}')

    @classmethod
    def _obs_df_reorder_columns(cls, obs_df):
        """Change order of columns of obs_df to the order specified in cls._OBS_DF_COLUMN_NAMES."""
        obs_df = obs_df[list(cls._OBS_DF_COLUMNS)]
        return obs_df

    @staticmethod
    def resolve_station_name(station_name_in):
        """Convert station name from Scintrex observation file
        (as Note) to the naming convention used in the output
        file (BEV conventions).

        Parameters
        ----------
        station_name_in : str
            Station name string as written to the observation file
            as note.

        Returns
        -------
        Corrected station name : str
        """
        station_name_in = station_name_in.upper()
        # Check first letter of name in order detect the station type:
        if station_name_in[0] == 'S':
            station_name_out = station_name_in
        elif station_name_in[0] == 'P':
            station_name_out = 'P  ' + station_name_in[1:]
        elif station_name_in[0] == 'T' and '.' in station_name_in:
            [str1_tmp, str2_tmp] = station_name_in.split('.')
            station_name_out = 'T{0:>4} {1:>3}'.format(str1_tmp[1:], str2_tmp)
        elif station_name_in[0] == 'N':
            station_name_out = station_name_in
        else:
            station_name_out = station_name_in.replace('.', '-')
        return station_name_out.upper()

    @staticmethod
    def get_dhb_dhf(dh_str):
        """Convert dhb and dhf from notes in the CG-5 observation file
        to an actual number. In the notes '.' is used instead of '-'.

        Parameters
        ----------
        dh_str : str
            Height difference [m] with '.' instead of '-'.

        Returns
        -------
        Height difference [m] : float
        """
        if dh_str.startswith('.'):
            dh_str = '-' + dh_str[1:]
        return float(dh_str)

    @property
    def number_of_observations(self):
        """Return the number of observations in the survey."""
        return len(self.obs_df)

    @property
    def number_of_setups(self):
        """Return the number of setups in the survey."""
        return len(set(self.obs_df['setup_id']))

    def __str__(self):
        if self.obs_df is None:
            return 'Empty CG-6 survey.'
        else:
            return f'Survey {self.survey_name} with {self.number_of_observations} observations in {self.number_of_setups} setups.'

    def plot_g_values(self, station_names=None):
        """Plot g-values of selected or all stations in the df.

        Notes
        -----
        This method requires matplotlib as optional dependency!

        Parameters
        ----------
        station_names : list of str, optional
            List of names of stations for which the observations will be plotted.
            The default value is None which implied that the data of all stations
            is plotted.

        """
        try:
            import matplotlib.pyplot as plt
        except ImportError:
            print('Package matplotlib not installed.')
            return

            # Check if obs data is available first:
        if self.obs_df is not None:
            # Get list of station names and loop over them:
            if station_names is None:
                station_names = self.obs_df['station'].unique()
            fig, ax = plt.subplots()
            for station_name in station_names:
                x = self.obs_df[self.obs_df['station'] == station_name].ref_time
                y = self.obs_df[self.obs_df['station'] == station_name].g_corr_mugal
                ax.scatter(x, y, label=station_name)
            ax.legend()
            ax.grid(True)
            ax.set_xlabel('Time')
            ax.set_ylabel('g [µGal]')
            ax.set_title('survey: ' + self.survey_name)

            # Data cursor:
            fig.canvas.mpl_connect('pick_event', DataCursor(plt.gca()))
            plt.show()


# Run as standalone program:
if __name__ == "__main__":
    s_solo = CG6Survey.from_cg6solo_file('../../data/CG6_data/Heligrav_2023/CG-6_0265_HELIGRAV_2023-09-11.dat',
                                         300,
                                         verbose=True)
    print(s_solo)
    s_solo.plot_g_values()
else:
    pass



"""Models for the calculation of atmospheric pressure corrections.

Copyright (C) 2023  Andreas Hellerschmied <andreas.hellerschmied@bev.gv.at>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
from typing import Tuple, Any

def normal_air_pressure_iso(height_m: float) -> float:
    """Returns the normal air pressure according to ISO 2533:1975.

    Parameters
    ----------
    height_m : float
        Physical height [m].

    Notes
    -----
    Reference:
    Wziontek et al. (2021): Status of the International Gravity Reference System and Frame, Journal of
    Geodesy, 95:7, https://doi.org/10.1007/s00190-020-01438-9
    Torge (1989): Gravimetry. De Gruyter, ISBN 3-11-010702-3, p. 323, Ch. 9.2.3

    Returns
    -------
    float
        Normal air pressure [hPa]

    """
    return 1013.25 * (1 - 0.0065 * height_m / 288.15) ** 5.2559


def pressure_correction_iso(height_m: float, p_obs_hpa: float, admittance: float = 0.3) -> tuple[float | Any, float]:
    """Returns the atmospheric pressure correction determined by the station height and ISO normal air pressure.

    Parameters
    ----------
    height_m : float
        Physical height [m].
    p_obs_hpa : float
        Observed air pressure [hPa]
    admittance : float, optional (default = 0.3)
        Admittance factor for the determination of pressure corrections based on the difference between measured and
        normal air pressure.

    Notes
    -----
    Reference:
    Wziontek et al. (2021): Status of the International Gravity Reference System and Frame, Journal of
    Geodesy, 95:7, https://doi.org/10.1007/s00190-020-01438-9
    Torge (1989): Gravimetry. De Gruyter, ISBN 3-11-010702-3, p. 323, Ch. 9.2.3

    Returns
    -------
    dg_mugal : float
        Atmopheric pressure correction [µGal].
    pn_hpa : float
        Normal air pressure according to ISO 2533:1975 [hPa].
    """
    pn_hpa = normal_air_pressure_iso(height_m)
    dp_hpa = p_obs_hpa - pn_hpa
    dg_mugal = admittance * dp_hpa
    return dg_mugal, pn_hpa


def pressure_correction_iso_pandas_series(height_m, p_obs_hpa, admittance: float = 0.3):
    """Returns the atmospheric pressure correction determined by the station height and ISO normal air pressure.

        Parameters
        ----------
        height_m : pandas Series of float values
            Physical height [m].
        p_obs_hpa : pandas Series of float values
            Observed air pressure [hPa]
        admittance : float, optional (default = 0.3)
            Admittance factor for the determination of pressure corrections based on the difference between measured and
            normal air pressure.

        Notes
        -----
        Reference:
        Wziontek et al. (2021): Status of the International Gravity Reference System and Frame, Journal of
        Geodesy, 95:7, https://doi.org/10.1007/s00190-020-01438-9
        Torge (1989): Gravimetry. De Gruyter, ISBN 3-11-010702-3, p. 323, Ch. 9.2.3

        Returns
        -------
        dg_mugal : float
            Atmospheric pressure correction [µGal].
        pn_hpa : float
            Normal air pressure according to ISO 2533:1975 [hPa]. `NaN` is returned for each item with missing observed
            pressure value in `p_obs_hpa`.
        """

    pn_hpa = height_m.apply(normal_air_pressure_iso)
    dp_hpa = p_obs_hpa - pn_hpa
    dg_mugal = admittance * dp_hpa
    return dg_mugal, pn_hpa

